"""
Install and manage chutes on the host.

Endpoints for these functions can be found under /api/v1/chutes.
"""

import json
import os
import re
import subprocess
import tarfile
import tempfile

import six

from autobahn.twisted.resource import WebSocketResource
from klein import Klein

from paradrop.base import pdutils, settings
from paradrop.base.output import out
from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.core.config import resource
from paradrop.core.container.chutecontainer import ChuteContainer
from paradrop.lib.utils import pdosq

from . import cors
from . import hostapd_control


class ChuteCacheEncoder(json.JSONEncoder):
    """
    JSON encoder for chute cache dictionary.

    The chute cache can contain arbitrary objects, some of which may not be
    JSON-serializable.  This encoder returns handles unserializable objects by
    returning the `repr` string.
    """
    def default(self, o):
        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return repr(o)


class ChuteEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return o.__dict__


class UpdateEncoder(json.JSONEncoder):
    def default(self, o):
        result = {
            'created': o.createdTime,
            'responses': o.responses,
            'failure': o.failure
        }
        return result


def tarfile_is_safe(tar):
    """
    Check the names of files in the archive for safety.

    Returns True if all paths are relative and safe or False if
    any of the paths are absolute (leading slash) or try to access
    parent directories (leading ..).
    """
    for member in tar:
        # normpath is useful here because it correctly normalizes "a/../../c"
        # to "../c".
        path = os.path.normpath(member.name)
        if os.path.isabs(path) or path.startswith(".."):
            return False
    return True


def extract_tarred_chute(data):
    tar = tarfile.TarFile(fileobj=data)
    if not tarfile_is_safe(tar):
        raise Exception("Tarfile contains unsafe paths")

    tempdir = tempfile.mkdtemp()
    tar.extractall(tempdir)

    configfile = os.path.join(tempdir, "paradrop.yaml")
    if not os.path.isfile(configfile):
        raise Exception("No paradrop.yaml file found in chute source")

    paradrop_yaml = pdosq.read_yaml_file(configfile, default={})

    return (tempdir, paradrop_yaml)


def chute_access_allowed(request, chute):
    # Allow caller to pass either chute object or the name of a chute.
    if isinstance(chute, six.string_types):
        chute = ChuteStorage.chuteList.get(chute, None)

    # All users are allowed to install a chute if it does not conflict with an
    # existing one.
    if chute is None:
        return True

    # Only an admin or a chute owner are allowed to modify an existing chute.
    if request.user == chute.get_owner() or request.user.role in ["owner", "admin"]:
        return True
    else:
        return False


def permission_denied(request):
    request.setResponseCode(403)
    return json.dumps({
        "error": "You do not have permission to access the requested chute."
    })


class ChuteApi(object):
    routes = Klein()

    def __init__(self, update_manager):
        self.update_manager = update_manager

    @routes.route('/', methods=['GET'])
    def get_chutes(self, request):
        """
        List installed chutes.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "environment": {},
               "name": "hello-world",
               "allocation": {
                 "cpu_shares": 1024,
                 "prioritize_traffic": false
               },
               "state": "running",
               "version": "x1511808778",
               "resources": null
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        chuteStorage = ChuteStorage()
        chutes = chuteStorage.getChuteList()
        allocation = resource.computeResourceAllocation(chutes)

        result = []
        for chute in chutes:
            service_info = {}
            for service in chute.get_services():
                container_name = service.get_container_name()
                container = ChuteContainer(container_name)

                service_info[service.name] = {
                    'allocation': allocation.get(container_name, None),
                    'state': container.getStatus()
                }

            chute_info = {
                'name': chute.name,
                'owner': chute.get_owner(),
                'state': chute.state,
                'services': service_info,
                'version': getattr(chute, 'version', None),
                'environment': getattr(chute, 'environment', None),
                'resources': getattr(chute, 'resources', None)
            }

            result.append(chute_info)

        return json.dumps(result, cls=ChuteEncoder)

    @routes.route('/', methods=['POST'])
    def create_chute(self, request):
        cors.config_cors(request)

        update = dict(updateClass='CHUTE',
                      updateType='create',
                      user=request.user,
                      tok=pdutils.timeint())

        ctype = request.requestHeaders.getRawHeaders('Content-Type',
                default=[None])[0]
        if ctype == "application/x-tar":
            workdir, paradrop_yaml = extract_tarred_chute(request.content)
            config = paradrop_yaml.get("config", {})

            # Try to read chute name from top level (preferred) or from config
            # object (deprecated).
            if 'name' in paradrop_yaml:
                update['name'] = paradrop_yaml['name']
            elif 'name' in config:
                out.warn("Deprecated: move chute name to top level of config file.")
                update['name'] = config['name']
            else:
                raise Exception("Chute name not found in configuration file.")

            update['workdir'] = workdir
            update.update(config)
        else:
            # TODO: this case is not tested
            body = json.loads(request.content.read())
            config = body['config']
            update.update(config)

        # Set a time-based version number for side-loaded chutes because we do
        # not expect the to receive it from the config file.
        update['version'] = "x{}".format(update['tok'])

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/<chute>', methods=['GET'])
    def get_chute(self, request, chute):
        """
        Get information about an installed chute.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/hello-world

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "environment": {},
             "name": "hello-world",
             "allocation": {
               "cpu_shares": 1024,
               "prioritize_traffic": false
             },
             "state": "running",
             "version": "x1511808778",
             "resources": null
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        if not chute_access_allowed(request, chute_obj):
            return permission_denied(request)

        chuteStorage = ChuteStorage()
        allocation = resource.computeResourceAllocation(
                chuteStorage.getChuteList())

        service_info = {}
        for service in chute_obj.get_services():
            container_name = service.get_container_name()
            container = ChuteContainer(container_name)

            service_info[service.name] = {
                'allocation': allocation.get(container_name, None),
                'state': container.getStatus()
            }

        result = {
            'name': chute,
            'state': chute_obj.state,
            'services': service_info,
            'version': getattr(chute_obj, 'version', None),
            'environment': getattr(chute_obj, 'environment', None),
            'resources': getattr(chute_obj, 'resources', None)
        }

        return json.dumps(result, cls=ChuteEncoder)

    @routes.route('/<chute>', methods=['PUT'])
    def update_chute(self, request, chute):
        cors.config_cors(request)

        if not chute_access_allowed(request, chute):
            return permission_denied(request)

        update = dict(updateClass='CHUTE',
                      updateType='update',
                      tok=pdutils.timeint(),
                      user=request.user,
                      name=chute)

        ctype = request.requestHeaders.getRawHeaders('Content-Type',
                default=[None])[0]
        if ctype == "application/x-tar":
            workdir, paradrop_yaml = extract_tarred_chute(request.content)
            config = paradrop_yaml.get("config", {})

            # Try to read chute name from top level (preferred) or from config
            # object (deprecated).
            if 'name' in paradrop_yaml:
                update['name'] = paradrop_yaml['name']
            elif 'name' in config:
                out.warn("Deprecated: move chute name to top level of config file.")
                update['name'] = config['name']
            else:
                raise Exception("Chute name not found in configuration file.")

            update['workdir'] = workdir
            update.update(config)
        else:
            body = json.loads(request.content.read())
            config = body['config']

            update.update(config)

        # Set a time-based version number for side-loaded chutes because we do
        # not expect the to receive it from the config file.
        update['version'] = "x{}".format(update['tok'])

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/<chute>', methods=['DELETE'])
    def delete_chute(self, request, chute):
        cors.config_cors(request)

        if not chute_access_allowed(request, chute):
            return permission_denied(request)

        update = dict(updateClass='CHUTE',
                      updateType='delete',
                      tok=pdutils.timeint(),
                      user=request.user,
                      name=chute)

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/<chute>/stop', methods=['POST'])
    def stop_chute(self, request, chute):
        cors.config_cors(request)

        if not chute_access_allowed(request, chute):
            return permission_denied(request)

        update = dict(updateClass='CHUTE',
                      updateType='stop',
                      tok=pdutils.timeint(),
                      user=request.user,
                      name=chute)

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/<chute>/start', methods=['POST'])
    def start_chute(self, request, chute):
        cors.config_cors(request)

        if not chute_access_allowed(request, chute):
            return permission_denied(request)

        update = dict(updateClass='CHUTE',
                      updateType='start',
                      tok=pdutils.timeint(),
                      user=request.user,
                      name=chute)

        try:
            body = json.loads(request.content.read())

            # Chute environment variables can be replaced during the operation.
            update['environment'] = body['environment']
        except Exception:
            pass

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/<chute>/restart', methods=['POST'])
    def restart_chute(self, request, chute):
        cors.config_cors(request)

        if not chute_access_allowed(request, chute):
            return permission_denied(request)

        update = dict(updateClass='CHUTE',
                      updateType='restart',
                      tok=pdutils.timeint(),
                      user=request.user,
                      name=chute)

        try:
            body = json.loads(request.content.read())

            # Chute environment variables can be replaced during the operation.
            update['environment'] = body['environment']
        except Exception:
            pass

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        request.setHeader('Content-Type', 'application/json')
        return json.dumps(result)

    @routes.route('/<chute>/cache', methods=['GET'])
    def get_chute_cache(self, request, chute):
        """
        Get chute cache contents.

        The chute cache is a key-value store used during chute installation.
        It can be useful for debugging the Paradrop platform.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)

            result = chute_obj.getCacheContents()
            return json.dumps(result, cls=ChuteCacheEncoder)
        except KeyError:
            request.setResponseCode(404)
            return "{}"

    @routes.route('/<chute>/config', methods=['GET'])
    def get_chute_config(self, request, chute):
        """
        Get current chute configuration.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/config

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "net": {
               "wifi": {
                 "dhcp": {
                   "lease": "1h",
                   "limit": 250,
                   "start": 3
                 },
                 "intfName": "wlan0",
                 "options": {
                   "isolate": True
                 },
                 "ssid": "Free WiFi",
                 "type": "wifi"
               }
             }
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            config = chute_obj.create_specification()
            return json.dumps(config)
        except KeyError:
            request.setResponseCode(404)
            return "{}"

    @routes.route('/<chute>/config', methods=['PUT'])
    def set_chute_config(self, request, chute):
        """
        Update the chute configuration and restart to apply changes.

        **Example request**:

        .. sourcecode:: http

           PUT /api/v1/chutes/captive-portal/config
           Content-Type: application/json

           {
             "net": {
               "wifi": {
                 "dhcp": {
                   "lease": "1h",
                   "limit": 250,
                   "start": 3
                 },
                 "intfName": "wlan0",
                 "options": {
                   "isolate": True
                 },
                 "ssid": "Better Free WiFi",
                 "type": "wifi"
               }
             }
           }

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "change_id": 1
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        if not chute_access_allowed(request, chute):
            return permission_denied(request)

        update = dict(updateClass='CHUTE',
                      updateType='restart',
                      tok=pdutils.timeint(),
                      user=request.user,
                      name=chute)

        try:
            body = json.loads(request.content.read())
            update.update(body)
        except Exception:
            pass

        # We will return the change ID to the caller for tracking and log
        # retrieval.
        update['change_id'] = self.update_manager.assign_change_id()

        self.update_manager.add_update(**update)

        result = {
            'change_id': update['change_id']
        }
        return json.dumps(result)

    @routes.route('/<chute>/networks', methods=['GET'])
    def get_networks(self, request, chute):
        """
        Get list of networks configured for the chute.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "interface": "wlan0",
               "type": "wifi",
               "name": "wifi"
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "[]"

        result = []
        for iface in networkInterfaces:
            data = {
                'name': iface['name'],
                'type': iface['type'],
                'interface': iface['internalIntf']
            }
            result.append(data)

        return json.dumps(result)

    @routes.route('/<chute>/networks/<network>', methods=['GET'])
    def get_network(self, request, chute, network):
        """
        Get information about a network configured for the chute.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks/wifi

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "interface": "wlan0",
             "type": "wifi",
             "name": "wifi"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        data = {}
        for iface in networkInterfaces:
            if iface['name'] != network:
                continue

            data = {
                'name': iface['name'],
                'type': iface['type'],
                'interface': iface['internalIntf']
            }

        return json.dumps(data)

    @routes.route('/<chute>/networks/<network>/leases', methods=['GET'])
    def get_leases(self, request, chute, network):
        """
        Get current list of DHCP leases for chute network.

        Returns a list of DHCP lease records with the following fields:

        expires
          lease expiration time (seconds since Unix epoch)
        mac_addr
          device MAC address
        ip_addr
          device IP address
        hostname
          name that the device reported
        client_id
          optional identifier supplied by device

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks/wifi/leases

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "client_id": "01:5c:59:48:7d:b9:e6",
               "expires": "1511816276",
               "ip_addr": "192.168.128.64",
               "mac_addr": "5c:59:48:7d:b9:e6",
               "hostname": "paradrops-iPod"
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            externalSystemDir = chute_obj.getCache('externalSystemDir')
        except KeyError:
            request.setResponseCode(404)
            return "[]"

        leasefile = 'dnsmasq-{}.leases'.format(network)
        path = os.path.join(externalSystemDir, leasefile)

        # The format of the dnsmasq leases file is one entry per line with
        # space-separated fields.
        keys = ['expires', 'mac_addr', 'ip_addr', 'hostname', 'client_id']

        try:
            leases = []
            with open(path, 'r') as source:
                for line in source:
                    parts = line.strip().split()
                    leases.append(dict(zip(keys, parts)))
            return json.dumps(leases)

        except IOError:
            # During chute uninstallation, there is a small window where the
            # chute still exists but the leases file has been removed.
            request.setResponseCode(404)
            return "[]"

    @routes.route('/<chute>/networks/<network>/ssid', methods=['GET'])
    def get_ssid(self, request, chute, network):
        """
        Get currently configured SSID for the chute network.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks/wifi/ssid

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "ssid": "Free WiFi",
             "bssid": "02:00:08:24:03:dd"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        address = os.path.join(settings.PDCONFD_WRITE_DIR, "hostapd", ifname)
        return hostapd_control.execute(address, command="GET_CONFIG")

    @routes.route('/<chute>/networks/<network>/ssid', methods=['PUT'])
    def set_ssid(self, request, chute, network):
        """
        Change the configured SSID for the chute network.

        The change will not persist after a reboot. If a persistent change is
        desired, you should update the chute configuration instead.

        **Example request**:

        .. sourcecode:: http

           PUT /api/v1/chutes/captive-portal/networks/wifi/ssid
           Content-Type: application/json

           {
             "ssid": "Best Free WiFi"
           }

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "message": "OK"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        body = json.loads(request.content.read())
        if "ssid" not in body:
            raise Exception("ssid required")

        command = "SET ssid {}".format(body['ssid'])
        address = os.path.join(settings.PDCONFD_WRITE_DIR, "hostapd", ifname)
        return hostapd_control.execute(address, command=command)

    @routes.route('/<chute>/networks/<network>/hostapd_status', methods=['GET'])
    def get_hostapd_status(self, request, chute, network):
        """
        Get low-level status information from the access point.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks/wifi/hostapd_status

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "olbc_ht": "0",
             "cac_time_left_seconds": "N/A",
             "num_sta_no_short_slot_time": "0",
             "olbc": "1",
             "num_sta_non_erp": "0",
             "ht_op_mode": "0x4",
             "state": "ENABLED",
             "num_sta_ht40_intolerant": "0",
             "channel": "11",
             "bssid[0]": "02:00:08:24:03:dd",
             "ieee80211n": "1",
             "cac_time_seconds": "0",
             "num_sta[0]": "1",
             "ieee80211ac": "0",
             "phy": "phy0",
             "num_sta_ht_no_gf": "1",
             "freq": "2462",
             "num_sta_ht_20_mhz": "1",
             "num_sta_no_short_preamble": "0",
             "secondary_channel": "0",
             "ssid[0]": "Free WiFi",
             "num_sta_no_ht": "0",
             "bss[0]": "vwlan7e1b"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        address = os.path.join(settings.PDCONFD_WRITE_DIR, "hostapd", ifname)
        return hostapd_control.execute(address, command="STATUS")

    @routes.route('/<chute>/networks/<network>/stations', methods=['GET'])
    def get_stations(self, request, chute, network):
        """
        Get detailed information about connected wireless stations.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks/wifi/stations

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "rx_packets": "230",
               "tdls_peer": "no",
               "authenticated": "yes",
               "rx_bytes": "12511",
               "tx_bitrate": "1.0 MBit/s",
               "tx_retries": "0",
               "signal": "-45 [-49, -48] dBm",
               "authorized": "yes",
               "rx_bitrate": "65.0 MBit/s MCS 7",
               "mfp": "no",
               "tx_failed": "0",
               "inactive_time": "4688 ms",
               "mac_addr": "5c:59:48:7d:b9:e6",
               "tx_bytes": "34176",
               "wmm_wme": "yes",
               "preamble": "short",
               "tx_packets": "88",
               "signal_avg": "-44 [-48, -47] dBm"
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "[]"

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        cmd = ['iw', 'dev', ifname, 'station', 'dump']
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        stations = []
        current = {}

        for line in proc.stdout:
            line = line.strip()

            match = re.match("Station\s+(\S+)\s+.*", line)
            if match is not None:
                current = {
                    'mac_addr': match.group(1)
                }
                stations.append(current)
                continue

            match = re.match("(.*):\s+(.*)", line)
            if match is not None:
                key = match.group(1).lower().replace(' ', '_').replace('/', '_')
                current[key] = match.group(2)

        return json.dumps(stations)

    @routes.route('/<chute>/networks/<network>/stations/<mac>', methods=['GET'])
    def get_station(self, request, chute, network, mac):
        """
        Get detailed information about a connected station.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/chutes/captive-portal/networks/wifi/stations/5c:59:48:7d:b9:e6

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "rx_packets": "230",
             "tdls_peer": "no",
             "authenticated": "yes",
             "rx_bytes": "12511",
             "tx_bitrate": "1.0 MBit/s",
             "tx_retries": "0",
             "signal": "-45 [-49, -48] dBm",
             "authorized": "yes",
             "rx_bitrate": "65.0 MBit/s MCS 7",
             "mfp": "no",
             "tx_failed": "0",
             "inactive_time": "4688 ms",
             "mac_addr": "5c:59:48:7d:b9:e6",
             "tx_bytes": "34176",
             "wmm_wme": "yes",
             "preamble": "short",
             "tx_packets": "88",
             "signal_avg": "-44 [-48, -47] dBm"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        cmd = ['iw', 'dev', ifname, 'station', 'get', mac]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        station = {}

        for line in proc.stdout:
            line = line.strip()

            match = re.match("Station\s+(\S+)\s+.*", line)
            if match is not None:
                station['mac_addr'] = match.group(1)
                continue

            match = re.match("(.*):\s+(.*)", line)
            if match is not None:
                key = match.group(1).lower().replace(' ', '_').replace('/', '_')
                station[key] = match.group(2)

        return json.dumps(station)

    @routes.route('/<chute>/networks/<network>/stations/<mac>', methods=['DELETE'])
    def delete_station(self, request, chute, network, mac):
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return "{}"

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        cmd = ['iw', 'dev', ifname, 'station', 'del', mac]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        messages = []
        for line in proc.stdout:
            line = line.strip()
            messages.append(line)

        return json.dumps(messages)

    @routes.route('/<chute>/networks/<network>/hostapd_control/ws', branch=True, methods=['GET'])
    def hostapd_control(self, request, chute, network):
        try:
            chute_obj = ChuteStorage.chuteList[chute]
            if not chute_access_allowed(request, chute_obj):
                return permission_denied(request)
            networkInterfaces = chute_obj.getCache('networkInterfaces')
        except KeyError:
            request.setResponseCode(404)
            return ""

        ifname = None
        for iface in networkInterfaces:
            if iface['name'] == network:
                ifname = iface['externalIntf']
                break

        ctrl_iface = os.path.join(settings.PDCONFD_WRITE_DIR, "hostapd", ifname)
        factory = hostapd_control.HostapdControlWSFactory(ctrl_iface)
        factory.setProtocolOptions(autoPingInterval=10, autoPingTimeout=5)
        return WebSocketResource(factory)



import json
import os

from autobahn.twisted.websocket import WebSocketServerFactory
from twisted.internet import interfaces, reactor
from twisted.internet.defer import Deferred
from twisted.internet.protocol import ConnectedDatagramProtocol
from zope.interface import implementer

from .generic_ws import ProducerConsumerWsProtocol


@implementer(interfaces.IConsumer)
@implementer(interfaces.IPushProducer)
class HostapdControlProtocol(ConnectedDatagramProtocol):
    """
    Protocol for hostapd control interface.

    There are two modes of operation depending on the command argument:
    1. If command is None, the protocol connects to the control interface and
    saves all detected events to the queue.
    2. If command is not None, the protocol connects to the control interface,
    sends the command, and only saves messages related to the command
    execution.

    Reference:
    https://w1.fi/wpa_supplicant/devel/ctrl_iface_page.html
    """
    # Number of instances created, used for creating unique paths.
    counter = 0

    def __init__(self, consumer=None, command=None, emit_json=True):
        self.consumer = consumer
        self.command = command
        self.emit_json = emit_json

        self.producer = None
        self.streaming = False

        self.connected = False
        self.paused = False

        self.instance = HostapdControlProtocol.counter
        HostapdControlProtocol.counter += 1

    def datagramReceived(self, datagram):
        output = datagram

        if self.emit_json:
            obj = {}
            # If the line contains any equal signs, interpret it as a list of
            # key-value pairs. Otherwise, interpret it as a single message, e.g.
            # "OK".
            if "=" in datagram:
                for line in datagram.split("\n"):
                    # There might be an equal sign in the value, so limit the
                    # number of splits to one.
                    words = line.split("=", 1)
                    if len(words) < 2:
                        continue
                    obj[words[0]] = words[1]
            else:
                obj['message'] = datagram.rstrip()
            output = json.dumps(obj)

        if self.consumer is not None:
            self.consumer.write(output)

    def startProtocol(self):
        self.connected = True
        if self.command is not None:
            self.transport.write(self.command)

    def bindAddress(self):
        """
        Choose a path for the client side of the UNIX socket.

        This should be unique to avoid errors.
        - Use the tmp directory, so that they are sure to be cleaned up.
        - Use the PID for uniqueness if the daemon is restarted.
        - Use the counter for uniqueness across multiple requests.
        """
        return "/tmp/hostapd-{}-{}.sock".format(os.getpid(), self.instance)

    #
    # IConsumer interface
    #

    def registerProducer(self, producer, streaming):
        self.producer = producer
        self.streaming = streaming

        # We are ready to receive data from the producer when the websocket
        # connection is open.
        if self.connected:
            producer.resumeProducing()
        else:
            producer.pauseProducing()

    def unregisterProducer(self):
        self.producer = None

    def write(self, data):
        self.transport.write(data)
        if not self.streaming:
            self.producer.resumeProducing()

    #
    # IPushProducer interface
    #

    def pauseProducing(self):
        self.paused = True
        if self.connected:
            self.transport.pauseProducing()

    def resumeProducing(self):
        self.paused = False
        if self.connected:
            self.transport.resumeProducing()

    def stopProducing(self):
        if self.connected:
            self.transport.loseConnection()


@implementer(interfaces.IConsumer)
class SingleItemConsumer(object):
    """
    Consumer that accepts a single item from a producer.

    This consumer waits for a single item from the producer and passes it to a
    Deferred object.
    """
    def __init__(self, deferred):
        self.deferred = deferred
        self.producer = None

    def registerProducer(self, producer, streaming):
        self.producer = producer
        producer.resumeProducing()

    def unregisterProducer(self):
        self.producer = None

    def write(self, data):
        self.deferred.callback(data)
        self.producer.stopProducing()


class HostapdControlWSFactory(WebSocketServerFactory):
    def __init__(self, control_interface):
        WebSocketServerFactory.__init__(self)
        self.control_interface = control_interface

    def buildProtocol(self, address):
        ws = ProducerConsumerWsProtocol()
        ws.factory = self

        control = HostapdControlProtocol(emit_json=False)
        reactor.connectUNIXDatagram(self.control_interface, control,
                bindAddress=control.bindAddress())

        ws.consumer = control
        control.consumer = ws

        ws.registerProducer(control, True)
        control.registerProducer(ws, True)

        return ws


def execute(address, command):
    deferred = Deferred()
    consumer = SingleItemConsumer(deferred)

    protocol = HostapdControlProtocol(consumer, command=command)
    consumer.registerProducer(protocol, True)

    reactor.connectUNIXDatagram(address, protocol, bindAddress=protocol.bindAddress())
    return deferred



###################################################################
# Copyright 2013-2018 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import attr

from paradrop.base.exceptions import ServiceNotFound


@attr.s
class Chute(object):
    """
    This Chute class provides the internal representation of a Paradrop chute.

    This class encapsulates the complex configuration details of a chute and
    provides a stable interface for the execution path even as the chute
    specification language evolves over time.

    The Chute class has minimal external dependencies, e.g. no dependency on
    the Docker API. Chute objects should be immutable, since they describe a
    desired software state at a fixed point in time.

    Args:
        name (str): The name of the chute.
        description (str): The human-friendly description of the chute.
        state (str): Desired run state of the chute ("running", "stopped").
        version (str): The version of the chute.
        config (dict): Configuration settings for the chute.
        environment (dict): Environment variables to set for all chute services.
    """
    STATE_INVALID = "invalid"
    STATE_DISABLED = "disabled"
    STATE_RUNNING = "running"
    STATE_FROZEN = "frozen"
    STATE_STOPPED = "stopped"

    description = attr.ib(default=None)
    name        = attr.ib(default=None)
    owner       = attr.ib(default=None)
    state       = attr.ib(default="running")
    version     = attr.ib(default=None)
    config      = attr.ib(default=attr.Factory(dict))
    environment = attr.ib(default=attr.Factory(dict))
    services    = attr.ib(default=attr.Factory(dict))
    web         = attr.ib(default=attr.Factory(dict))

    # The cache as a working storage of intermediate values has been moved
    # to the update object. Here, we set the cache right before saving the
    # chute to disk so that the values can be retrieved with the chute
    # list.
    _cache      = attr.ib(default=attr.Factory(dict))

    def __repr__(self):
        return "<Chute {} - {}>".format(self.name, self.state)

    def __str__(self):
        return "Chute:{}".format(self.name)

    def isRunning(self):
        """
        Check if the chute is supposed to be running.
        """
        return self.state == Chute.STATE_RUNNING

    def isValid(self):
        """Return True only if the Chute object we have has all the proper things defined to be in a valid state."""
        if (not self.name or len(self.name) == 0):
            return False
        return True

    def getCache(self, key):
        """
        Get a value from the cache or None if it does not exist.
        """
        return self._cache.get(key, None)

    def getCacheContents(self):
        """
        Get the contents of the cache as a dictionary.
        """
        return self._cache

    def setCache(self, key, value):
        """
        Set a value in the cache.

        Deprecated: Most of the cache functionality has been moved to the
        Update object because they are values that are used as temporary
        storage between one update step and the following steps. However, there
        are a few instances of cache values that we do still read from chute
        storage. Any calls to the getCache method throughout the project are
        still depending on this functionality, so we have corresponding calls
        to setCache that ensure the required information is present in the
        chute cache and not just in the update cache. Eventually, we should
        remove this dependency either by using a less stateful design or by
        formalizing the process for storing persistent chute state, such as the
        networkInterfaces list.
        """
        self._cache[key] = value

    def updateCache(self, other):
        """
        Update the chute cache from another dictionary.
        """
        self._cache.update(other)

    def getConfiguration(self):
        """
        Get the chute's configuration object.
        """
        # TODO: Split metadata (e.g. name and version) from configuration data.
        # Currently, we do this by selectively copying from __dict__.  A
        # cleaner separation would require refactoring all the way through how
        # we create update objects.
        return self.config

    def getHostConfig(self):
        """
        Get the chute's host_config options for Docker.

        Returns an empty dictionary if there is no host_config setting.
        """
        return self.config.get('host_config', {})

    def getWebPort(self):
        """
        Get the port configured for the chute's web server.

        Returns port (int) or None if no port is configured.
        """
        try:
            return int(self.config['web']['port'])
        except:
            return None

    def inherit_attributes(self, other):
        """
        Inherit attributes from another version of the chute.

        If any settings are None or missing in this chute but present in the
        other version, they will be copied over. The return value is a
        dictionary containing changes that were applied.
        """
        changes = {}
        for attr_name in ["name", "description", "state", "version"]:
            if getattr(self, attr_name, None) is None:
                other_value = getattr(other, attr_name, None)
                if other_value is not None:
                    setattr(self, attr_name, other_value)
                    changes[attr_name] = other_value
        return changes

    def add_service(self, service):
        """
        Add a service to the chute.
        """
        self.services[service.name] = service

    def create_specification(self):
        """
        Create a new chute specification from the existing chute.

        This is a completely clean copy of all information necessary to rebuild
        the Chute object. It should contain only primitive types, which can
        easily be serialized as JSON or YAML.
        """
        def no_privates(a, _):
            return not a.name.startswith('_')

        return attr.asdict(self, filter=no_privates)

    def get_default_service(self):
        """
        Get one of the chute's services designated as the default one.

        This is more for convenience with existing API functions where the
        caller did not need to specify a service because prior to 0.12.0,
        chutes could only have one Docker container. We use some heuristics
        such as the service's name is "main" to identify one of the services as
        the default.
        """
        if "main" in self.services:
            return self.services['main']

        # Sort by name and return the first one.
        name = min(self.services)
        return self.services[name]

    def get_environment(self):
        """
        Get the chute environment variables.

        These are defined by the developer or administrator and passed to all
        services that belong to the chute.
        """
        return getattr(self, "environment", {})

    def get_owner(self):
        """
        Get the name of the user who owns this installed chute.
        """
        return getattr(self, "owner", None)

    def get_service(self, name):
        """
        Get a service by name.
        """
        if name in self.services:
            return self.services[name]
        else:
            raise ServiceNotFound(name)

    def get_services(self):
        """
        Get a list of services installed by this chute.
        """
        return self.services.values()

    def get_web_port_and_service(self):
        """
        Get the port and Service object that provides this chutes web service.

        Returns a tuple containing the port number and Service object. Both
        values will be None if a web service is not configured.
        """
        port = self.web.get("port", None)
        if port is None:
            return None, None

        name = self.web.get("service", None)
        return port, self.get_service(name)



import netifaces
import ipaddress
import six

from builtins import str

from paradrop.base.output import out
from paradrop.base import constants, pdutils, settings
from paradrop.lib.utils import datastruct, uci

from . import uciutils

# TODO: Instead of being a constant, look at device capabilities.
MAX_AP_INTERFACES = 8

# We use a 4-digit hex string, which gives us this maximum.
MAX_INTERFACE_NUMBERS = 65536

# Size of subnets to assign to chutes.
SUBNET_SIZE = 24

# Extra Wi-Fi interface options that we will pass on directly to confd.
# Deprecated: move to 'options' object after 1.0 release.
IFACE_EXTRA_OPTIONS = set([
    "auth_server", "auth_secret", "acct_server", "acct_secret", "acct_interval"
])


def reclaimNetworkResources(chute):
    """
    Reclaim network resources for a previously running chute.

    This function only applies to the special case in which pd starts up and
    loads a list of chutes that were running.  This function marks their IP
    addresses and interface names as taken so that new chutes will not use the
    same values.
    """
    pass


def split_interface_type(itype):
    words = itype.split("-")
    if len(words) == 1:
        return words[0], None
    else:
        return tuple(words)


def getWifiKeySettings(cfg, iface):
    """
    Read encryption settings from cfg and transfer them to iface.
    """
    # If 'key' is present, but 'encryption' is not, then default to
    # psk2.  If 'key' is not present, and 'encryption' is psk or psk2,
    # then we have an error.
    iface['encryption'] = "none"
    if 'key' in cfg:
        iface['key'] = cfg['key']
        iface['encryption'] = 'psk2'  # default to psk2
    if 'encryption' in cfg:
        iface['encryption'] = cfg['encryption']
        if cfg['encryption'].startswith('psk') and 'key' not in cfg:
            out.warn("Key field must be defined "
                     "when encryption is enabled.")
            raise Exception("No key field defined for WiFi encryption")


def select_chute_subnet_pool(host_config):
    """
    Select IP subnet to use as pool for chutes.

    Behavior depends on whether a static subnet is configured or auto
    configuration is requested. If the chuteSubnetPool option is set to 'auto',
    then we check the WAN interface address and choose either 10.128.0.0/9 or
    192.168.128.0/17 to avoid conflict.  Otherwise, we used the specified
    subnet.
    """
    pool = datastruct.getValue(host_config, "system.chuteSubnetPool", 'auto')
    wan_ifname = datastruct.getValue(host_config, 'wan.interface', 'eth0')

    if pool == 'auto':
        addresses = netifaces.ifaddresses(wan_ifname)
        ipv4_addrs = addresses.get(netifaces.AF_INET, [])

        if any(x['addr'].startswith("10.") for x in ipv4_addrs):
            return "192.168.128.0/17"
        else:
            return "10.128.0.0/9"

    else:
        return pool


def chooseSubnet(update, cfg, iface):
    reservations = update.cache_get('subnetReservations')

    # Check if the chute configuration requests a specific network.
    #
    # TODO: Implement a mode where a chute can request a preferred network, but
    # do not fail if it is unavailable.
    requested = cfg.get('ipv4_network', None)
    if requested is not None:
        network = ipaddress.ip_network(str(requested))
        if network in reservations:
            raise Exception(
                "Could not assign network {}, network already in use".format(
                    requested))
        else:
            reservations.add(network)
            return network

    # Get subnet configuration settings from host configuration.
    host_config = update.cache_get('hostConfig')
    network_pool = select_chute_subnet_pool(host_config)
    prefix_size = datastruct.getValue(host_config, "system.chutePrefixSize",
                                      SUBNET_SIZE)

    network = ipaddress.ip_network(str(network_pool))
    if prefix_size < network.prefixlen:
        raise Exception(
            "Router misconfigured: prefix size {} is invalid for network {}".
            format(prefix_size, network))

    subnets = network.subnets(new_prefix=prefix_size)
    for subnet in subnets:
        if subnet not in reservations:
            reservations.add(subnet)
            return subnet

    raise Exception("Could not find an available subnet")


def chooseExternalIntf(update, iface):
    if iface['type'] == "wifi-ap":
        # Generate initial portion (prefix) of interface name.
        #
        # NOTE: We add a "v" in front of the interface name to avoid triggering
        # the udev persistent net naming rules, which are hard-coded to certain
        # typical strings such as "eth*" and "wlan*" but not "veth*" or
        # "vwlan*".  We do NOT want udev renaming our virtual interfaces.
        prefix = "vwlan"

        # This name should be unique on the system, so the hash is very unlikely to
        # collide with anything.  It still can collide, but this will be our first
        # choice for an interface number.
        name = "{}:{}".format(update.new.name, iface['internalIntf'])
        base = hash(name)

        reservations = update.cache_get('interfaceReservations')

        for i in range(MAX_INTERFACE_NUMBERS):
            number = (base + i) % MAX_INTERFACE_NUMBERS
            intf = "{}{:04x}".format(prefix, number)
            if intf not in reservations:
                reservations.add(intf)
                return intf

        raise Exception("Could not find an available interface name")

    else:
        # For monitor and managed mode, use the existing primary interface.
        devices = update.cache_get('networkDevicesByName')
        if iface['device'] not in devices:
            raise Exception("Could not find device {}".format(iface['device']))

        device = devices[iface['device']]
        if 'primary_interface' not in device:
            raise Exception("Primary interface for {} not detected".format(
                device['name']))

        return device['primary_interface']


def getInterfaceAddress(update, name, cfg, iface):
    """
    Dynamically select IP address for the chute interface.

    This function will use a subnet from the chute subnet pool and assign IP
    addresses to the external (in host) and internal (in chute) interfaces.

    The addresses are stored in the iface object.
    """
    # Claim a subnet for this interface from the pool.
    subnet = chooseSubnet(update, cfg, iface)

    # Generate internal (in the chute) and external (in the host)
    # addresses.
    #
    # Example:
    # subnet: 192.168.30.0/24
    # netmask: 255.255.255.0
    # external: 192.168.30.1
    # internal: 192.168.30.2
    hosts = subnet.hosts()
    iface['subnet'] = subnet
    iface['netmask'] = str(subnet.netmask)
    iface['externalIpaddr'] = str(next(hosts))
    iface['internalIpaddr'] = str(next(hosts))

    # Generate the internal IP address with prefix length (x.x.x.x/y) for
    # convenience of other code that expect that format (e.g. pipework).
    iface['ipaddrWithPrefix'] = "{}/{}".format(iface['internalIpaddr'],
                                               subnet.prefixlen)


def getNetworkConfigWifi(update, name, cfg, iface):
    dtype, mode = split_interface_type(cfg['type'])

    # Later code still depends on the mode field.
    iface['mode'] = mode

    # Generate a name for the new interface in the host.
    iface['externalIntf'] = chooseExternalIntf(update, iface)
    if len(iface['externalIntf']) > constants.MAX_INTERFACE_NAME_LEN:
        out.warn("Interface name ({}) is too long\n".format(
            iface['externalIntf']))
        raise Exception("Interface name is too long")

    wireless = cfg.get("wireless", {})
    iface['wireless'] = wireless

    if iface['type'] in ["wifi-ap", "wifi-sta"]:
        # Check for required fields.
        res = pdutils.check(wireless, dict, ['ssid'])
        if res:
            out.warn('WiFi network interface definition {}\n'.format(res))
            raise Exception("Interface definition missing field(s)")

        iface['ssid'] = wireless['ssid']

    # Optional encryption settings
    getWifiKeySettings(wireless, iface)

    # Give a warning if the dhcp block is missing, since it is likely
    # that developers will want a DHCP server to go with their AP.
    if 'dhcp' not in cfg:
        out.warn("No dhcp block found for interface {}; "
                 "will not run a DHCP server".format(name))


def getNetworkConfigVlan(update, name, cfg, iface):
    res = pdutils.check(cfg, dict, ['vlan_id'])
    if res:
        raise Exception("Interface definition missing field(s)")

    # Generate a name for the new interface in the host.
    iface['externalIntf'] = "br-lan.{}".format(cfg['vlan_id'])

    # Prevent multiple chutes from using the same VLAN.
    reservations = update.cache_get('interfaceReservations')
    if iface['externalIntf'] in reservations:
        raise Exception("Interface {} already in use".format(
            iface['externalIntf']))
    reservations.add(iface['externalIntf'])


def getNetworkConfigLan(update, name, cfg, iface):
    # Claim a subnet for this interface from the pool.
    iface['externalIntf'] = iface['device']


def get_current_phy_conf(update, device_id):
    """
    Lookup current configuration for a network device.

    This includes information such as the Wi-Fi channel.

    Returns a dictionary, which may be empty if no configuration was found.
    """
    hostConfig = update.cache_get('hostConfig')

    for dev in hostConfig.get('wifi', []):
        if dev.get('id', None) == device_id:
            return dev

    return {}


def satisfies_requirements(obj, requirements):
    """
    Checks that an object satifies given requirements.

    Every key-value pair in the requirements object must be present in the
    target object for it to be considered satisfied.

    Returns True/False.
    """
    for key, value in six.iteritems(requirements):
        if key not in obj:
            return False
        if obj[key] != value:
            return False
    return True


def fulfillDeviceRequest(update, cfg, devices):
    """
    Find a physical device that matches the requested device type.

    Raises an exception if one cannot be found.
    """
    dtype, mode = split_interface_type(cfg['type'])

    # Get list of devices by requested type.
    devlist = devices.get(dtype, [])

    reservations = update.cache_get('deviceReservations')

    bestDevice = None
    bestScore = -1

    for device in devlist:
        dname = device['name']

        # If the configuration object includes a 'requests' object, then look
        # up configuration of the device and check additional constraints, e.g.
        # channel or hwmode.
        if 'requests' in cfg:
            phyconf = get_current_phy_conf(update, device['id'])
            if not satisfies_requirements(phyconf, cfg['requests']):
                continue

        # Monitor, station, and airshark mode interfaces require exclusive
        # access to the device.
        if dtype == "wifi" and mode in ["monitor", "sta", "airshark"]:
            if reservations[dname].count() > 0:
                continue

            # Choose the first one that matches.
            bestDevice = device
            break

        # AP mode interfaces can share a device, but not with monitor mode or
        # station mode.
        elif dtype == "wifi" and mode == "ap":
            if reservations[dname].count(mode="monitor") > 0:
                continue
            if reservations[dname].count(mode="sta") > 0:
                continue

            apcount = reservations[dname].count(mode="ap")

            # Avoid exceeding the max. number of AP interfaces.
            if apcount >= MAX_AP_INTERFACES:
                continue

            # Otherwise, prefer interfaces that have at least one AP already.
            # This preference leaves interfaces available for other purposes
            # (e.g. monitor mode).
            if apcount > bestScore:
                bestDevice = device
                bestScore = apcount

        else:
            # Handle other devices types, namely "lan".  Assume they require
            # exclusive access.
            if reservations[dname].count() > 0:
                continue
            else:
                bestDevice = device
                break

    if bestDevice is not None:
        out.info("Assign device {} for requested type {}".format(
            bestDevice['name'], dtype))
        reservations[bestDevice['name']].add(update.new.name, dtype, mode)
        return bestDevice

    raise Exception(
        "Could not satisfy requirement for device of type {}.".format(dtype))


def getExtraOptions(cfg):
    """
    Get dictionary of extra wifi-iface options that we are not interpreting but
    just passing on to pdconf.
    """
    # Using an 'options' object in the configuration is more extensible than
    # enumerating the possible extra options.
    options = cfg.get('options', {})

    # Deprecated: remove after 1.0 release.
    for key, value in six.iteritems(cfg):
        if key in IFACE_EXTRA_OPTIONS:
            options[key] = value

    return options


def getNetworkConfig(update):
    """
    For the Chute provided, return the dict object of a 100% filled out
    configuration set of network configuration. This would include determining
    what the IP addresses, interfaces names, etc...

    Store configuration in networkInterfaces cache entry.
    """

    # Notes:
    #
    # Fill in the gaps of knowledge between what the dev provided us in their
    # config and what would be required to add all the config changes to get
    # the chute working By the end of this function there should be a
    # cache:networkInterfaces key containing a list of dictionary objects for
    # each interface we will need to create, including netmasks IP addresses,
    # etc.. this is important to allow all other modules to know what the IP
    # addr is for this chute, etc..
    #
    # old code under lib.internal.chs.chutelxc same function name

    interfaces = list()

    # Put the list in the cache now (shared reference), so that if we fail out
    # of this function after partial completion, the abort function can take
    # care of what made it into the list.
    update.cache_set('networkInterfaces', interfaces)

    # Deprecated: we set this value in the chute cache here because there is
    # still code that depends on reading this from the chute storage.  This can
    # be removed when the corresponding call to getCache is removed.
    update.new.setCache('networkInterfaces', interfaces)

    # Make sure we only assign interfaces to running chutes.
    if not update.new.isRunning():
        return None

    devices = update.cache_get('networkDevices')

    for service in update.new.get_services():
        for name, cfg in six.iteritems(service.interfaces):
            # Check for required fields.
            res = pdutils.check(cfg, dict, ['type'])
            if res:
                out.warn('Network interface definition {}\n'.format(res))
                raise Exception("Interface definition missing field(s)")

            iface = {
                'name': name,  # Name (not used?)
                'service': service.name,  # Service name
                'type': cfg['type'],  # Type (wan, lan, wifi)
                'internalIntf': cfg['intfName'],  # Interface name in chute
                'l3bridge': cfg.get('l3bridge', None)  # Optional
            }

            getInterfaceAddress(update, name, cfg, iface)

            if cfg['type'].startswith("wifi"):
                # Try to find a physical device of the requested type.
                #
                # Note: we try this first because it can fail, and then we will not try
                # to allocate any resources for it.
                device = fulfillDeviceRequest(update, cfg, devices)
                iface['device'] = device['name']
                iface['phy'] = device['phy']

                getNetworkConfigWifi(update, name, cfg, iface)

            elif cfg['type'] == "vlan":
                getNetworkConfigVlan(update, name, cfg, iface)

            elif cfg['type'] == "lan":
                device = fulfillDeviceRequest(update, cfg, devices)
                iface['device'] = device['name']

                getNetworkConfigLan(update, name, cfg, iface)

            else:
                raise Exception("Unsupported network type, {}".format(
                    cfg['type']))

            # Pass on DHCP configuration if it exists.
            if 'dhcp' in cfg:
                iface['dhcp'] = cfg['dhcp']

            # TODO: Refactor!  The problem here is that `cfg` contains a mixture of
            # fields, some that we will interpret and some that we will pass on to
            # pdconf.  The result is a lot of logic that tests and copies without
            # actually accomplishing much.
            iface['options'] = getExtraOptions(cfg)

            interfaces.append(iface)

    update.cache_set('networkInterfaces', interfaces)

    # Deprecated: we set this value in the chute cache here because there is
    # still code that depends on reading this from the chute storage.  This can
    # be removed when the corresponding call to getCache is removed.
    update.new.setCache('networkInterfaces', interfaces)


def abortNetworkConfig(update):
    """
    Release resources claimed by chute network configuration.
    """
    pass


def getOSNetworkConfig(update):
    """
    Takes the network interface obj created by
    NetworkManager.getNetworkConfiguration and returns a properly formatted
    object to be passed to the UCIConfig class.  The object returned is a
    list of tuples (config, options).
    """
    # Notes:
    #
    # Take the cache:networkInterfaces key and generate the UCI specific OS
    # config we need to inact the chute changes, should create a
    # cache:osNetworkConfig key
    #
    # old code under lib.internal.chs.chutelxc same function name

    interfaces = update.cache_get('networkInterfaces')

    osNetwork = list()

    for iface in interfaces:
        config = {'type': 'interface', 'name': iface['externalIntf']}

        if iface['type'] == "wifi-monitor":
            if not settings.ALLOW_MONITOR_MODE:
                raise Exception("Monitor mode interfaces are disallowed by the system settings.")

            # Monitor mode is a special case - do not configure an IP address
            # for it.
            options = {'proto': 'none', 'ifname': iface['externalIntf']}

        else:
            # TODO: Check if configured IP addresses conflict with reservations
            # by chutes and fail a sethostconfig operation in that case.
            options = {
                'proto': 'static',
                'ipaddr': iface['externalIpaddr'],
                'netmask': iface['netmask'],
                'ifname': iface['externalIntf']
            }

        # Add to our OS Network
        osNetwork.append((config, options))

    update.cache_set('osNetworkConfig', osNetwork)


def setOSNetworkConfig(update):
    """
    Takes a list of tuples (config, opts) and saves it to the network config
    file.
    """

    # Notes:
    #
    # Takes the config generated (cache:osNetworkConfig) and uses the UCI
    # module to actually push it to the disk
    #
    # old code under lib.internal.chs.chutelxc same function name

    uciutils.setConfig(
        update,
        cacheKeys=['osNetworkConfig'],
        filepath=uci.getSystemPath("network"))


def revert_os_network_config(update):
    uciutils.setConfig(
        update,
        cacheKeys=['osNetworkConfig'],
        filepath=uci.getSystemPath("network"))


def getL3BridgeConfig(update):
    """
    Creates configuration sections for layer 3 bridging.
    """
    interfaces = update.cache_get('networkInterfaces')

    parprouted = list()

    for iface in interfaces:
        l3bridge = iface.get('l3bridge', None)
        if l3bridge is None:
            continue

        config = {'type': 'bridge'}
        options = {'interfaces': [iface['externalIntf'], l3bridge]}
        parprouted.append((config, options))

    update.cache_set('parproutedConfig', parprouted)


def setL3BridgeConfig(update):
    """
    Apply configuration for layer 3 bridging.
    """
    uciutils.setConfig(
        update,
        cacheKeys=['parproutedConfig'],
        filepath=uci.getSystemPath("parprouted"))


def revert_l3_bridge_config(update):
    uciutils.setConfig(
        update,
        cacheKeys=['parproutedConfig'],
        filepath=uci.getSystemPath("parprouted"))



#
# This module collects information about the current state of the system
# (chutes installed, hardware available, etc.) and reports that to the Paradrop
# server for remote management purposes.
#

import json
import os
import time

from twisted.internet import reactor

from paradrop.base.output import out
from paradrop.base import nexus, settings
from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.core.config import devices, hostconfig, resource, zerotier
from paradrop.core.container.chutecontainer import ChuteContainer
from paradrop.core.agent.http import PDServerRequest
from paradrop.core.system import system_info
from paradrop.core.system.system_status import SystemStatus
from paradrop.lib.misc.governor import GovernorClient


class StateReport(object):
    def __init__(self):
        # Record timestamp when report was created in case the server receives
        # multiple.
        self.timestamp = time.time()

        self.name = None
        self.osVersion = None
        self.paradropVersion = None
        self.pdinstallVersion = None
        self.chutes = []
        self.devices = []
        self.hostConfig = {}
        self.snaps = []
        self.zerotierAddress = None
        self.dmi = {}
        self.status = {}

    def toJSON(self):
        return json.dumps(self.__dict__)


class StateReportBuilder(object):
    def prepare(self):
        report = StateReport()

        report.name = nexus.core.info.pdid

        report.osVersion = system_info.getOSVersion()

        # We can get the paradrop version from the installed python package.
        report.paradropVersion = system_info.getPackageVersion('paradrop')

        report.chutes = []
        chuteStore = ChuteStorage()
        chutes = chuteStore.getChuteList()
        allocation = resource.computeResourceAllocation(chutes)

        for chute in chutes:
            service_info = {}
            for service in chute.get_services():
                container_name = service.get_container_name()
                container = ChuteContainer(container_name)

                service_info[service.name] = {
                    'allocation': allocation.get(container_name, None),
                    'state': container.getStatus()
                }

            # Use the default service (e.g. "main") to report the chute's
            # current state.
            service = chute.get_default_service()
            container = ChuteContainer(service.get_container_name())

            report.chutes.append({
                'name': chute.name,
                'desired': chute.state,
                'state': container.getStatus(),
                'services': service_info,
                'warning': None,
                'version': getattr(chute, 'version', None),
                'allocation': None,  # deprecated
                'environment': getattr(chute, 'environment', None),
                'external': getattr(chute, 'external', None),
                'resources': getattr(chute, 'resources', None)
            })

        report.devices = devices.listSystemDevices()
        report.hostConfig = hostconfig.prepareHostConfig(write=False)

        if GovernorClient.isAvailable():
            client = GovernorClient()
            report.snaps = client.listSnaps()

        report.zerotierAddress = zerotier.getAddress()
        report.dmi = system_info.getDMI()

        # Add CPU, memory, disk, and network interface information.  This gives
        # the controller useful debugging information such as high memory or
        # disk utilization and IP addresses.
        status_source = SystemStatus()
        report.status = status_source.getStatus(max_age=None)

        return report.__dict__


class TelemetryReportBuilder(object):
    def prepare(self):
        chuteStore = ChuteStorage()
        chutes = chuteStore.getChuteList()

        # All network interfaces: we will divide these into chute-specific
        # interfaces and system-wide interfaces.
        network = SystemStatus.getNetworkInfo()
        system_interfaces = set(network.keys())

        report = {
            'chutes': [],
            'network': [],
            'system': SystemStatus.getSystemInfo(),
            'time': time.time()
        }

        for chute in chutes:
            container = ChuteContainer(chute.name)

            chute_info = {
                'name': chute.name,
                'state': container.getStatus(),
                'network': []
            }

            try:
                pid = container.getPID()
                chute_info['process'] = SystemStatus.getProcessInfo(pid)
            except Exception:
                chute_info['process'] = None

            interfaces = chute.getCache('networkInterfaces')
            for iface in interfaces:
                ifname = iface['externalIntf']
                if ifname in network:
                    ifinfo = network[ifname]
                    ifinfo['name'] = ifname
                    ifinfo['type'] = iface.get('type', 'wifi')
                    chute_info['network'].append(ifinfo)
                    system_interfaces.remove(ifname)

            report['chutes'].append(chute_info)

        for ifname in system_interfaces:
            ifinfo = network[ifname]
            ifinfo['name'] = ifname
            ifinfo['type'] = None
            report['network'].append(ifinfo)

        return report


class ReportSender(object):
    def __init__(self, model="states", max_retries=None):
        self.max_retries = max_retries
        self.model = model
        self.retries = 0
        self.retryDelay = 1
        self.maxRetryDelay = 300

    def increaseDelay(self):
        self.retryDelay *= 2
        if self.retryDelay > self.maxRetryDelay:
            self.retryDelay = self.maxRetryDelay

    def send(self, report):
        request = PDServerRequest('/api/routers/{router_id}/' + self.model)
        d = request.post(**report)

        # Check for error code and retry.
        def cbresponse(response):
            if not response.success:
                out.warn('{} to {} returned code {}'.format(request.method,
                    request.url, response.code))
                if self.max_retries is None or self.retries < self.max_retries:
                    reactor.callLater(self.retryDelay, self.send, report)
                    self.retries += 1
                    self.increaseDelay()
                nexus.core.jwt_valid = False
            else:
                nexus.core.jwt_valid = True

        # Check for connection failures and retry.
        def cberror(ignored):
            out.warn('{} to {} failed'.format(request.method, request.url))
            if self.max_retries is None or self.retries < self.max_retries:
                reactor.callLater(self.retryDelay, self.send, report)
                self.retries += 1
                self.increaseDelay()
            nexus.core.jwt_valid = False

        d.addCallback(cbresponse)
        d.addErrback(cberror)
        return d


class NodeIdentitySender(ReportSender):
    def send(self, report):
        request = PDServerRequest('/api/routers/{router_id}')
        d = request.patch(*report)

        # Check for error code and retry.
        def cbresponse(response):
            if not response.success:
                out.warn('{} to {} returned code {}'.format(request.method,
                    request.url, response.code))
                if self.max_retries is None or self.retries < self.max_retries:
                    reactor.callLater(self.retryDelay, self.send, report)
                    self.retries += 1
                    self.increaseDelay()
                nexus.core.jwt_valid = False
            else:
                nexus.core.jwt_valid = True

        # Check for connection failures and retry.
        def cberror(ignored):
            out.warn('{} to {} failed'.format(request.method, request.url))
            if self.max_retries is None or self.retries < self.max_retries:
                reactor.callLater(self.retryDelay, self.send, report)
                self.retries += 1
                self.increaseDelay()
            nexus.core.jwt_valid = False

        d.addCallback(cbresponse)
        d.addErrback(cberror)
        return d


def sendNodeIdentity():
    path = os.path.join(settings.KEY_DIR, "node.pub")
    with open(path, "r") as source:
        public_key = source.read().strip()

    report = [
        {"op": "add", "path": "/ssh_public_key", "value": public_key}
    ]
    sender = NodeIdentitySender()
    return sender.send(report)


def sendStateReport():
    builder = StateReportBuilder()
    report = builder.prepare()

    sender = ReportSender()
    return sender.send(report)


def sendTelemetryReport():
    # Do not try to send telemetry report if not provisioned.
    if not nexus.core.provisioned():
        if not getattr(sendTelemetryReport, 'provisionWarningShown', False):
            out.warn("Unable to send telemetry report: not provisioned")
            sendTelemetryReport.provisionWarningShown = True
        return None

    builder = TelemetryReportBuilder()
    report = builder.prepare()

    sender = ReportSender(model="telemetry", max_retries=0)
    return sender.send(report)



"""
WebSocket session for communication with cloud controller.
"""
from __future__ import print_function
from autobahn.twisted.websocket import WebSocketClientFactory, WebSocketClientProtol

from paradrop.base.output import out
from paradrop.base import nexus


class HostWebSocketProtocol(WebSocketClientProtol):
    # Make this a class variable because a new WampSession object is created
    # whenever the WAMP connection resets, but we only call set_update_fetcher
    # on program initialization.
    update_fetcher = None

    def onConnect(self, response):
        out.info("WebSocket connected to {}".format(response.peer))

    def onOpen(self):
        out.info("WebSocket connection open.")
        self.sendMessage("subscribe:update:" + nexus.core.info.pdid)

    def onMessage(self, payload, isBinary):
        out.info("WebSocket message: {}".format(payload))

    def onClose(self, wasClean, code, reason):
        print("WebSocket connection closed: {}".format(reason))


class HostWebSocketFactory(WebSocketClientFactory):
    pass


# class WampSession(BaseSession):
#
#    def __init__(self, *args, **kwargs):
#        self.uriPrefix = 'org.paradrop.'
#        super(WampSession, self).__init__(*args, **kwargs)
#
#    def onConnect(self):
#        out.info('Router session connected.')
#        if (nexus.core.info.pdid and nexus.core.getKey('apitoken')):
#            out.info('Starting WAMP-CRA authentication on realm "{}" as user "{}"...'\
#                     .format(self.config.realm, nexus.core.info.pdid))
#            self.join(self.config.realm, [u'wampcra'],
#                    six.u(nexus.core.info.pdid))
#
#    def onChallenge(self, challenge):
#        if challenge.method == u"wampcra":
#            out.verbose("WAMP-CRA challenge received: {}".format(challenge))
#
#            wampPassword = nexus.core.getKey('apitoken')
#            if u'salt' in challenge.extra:
#                # salted secret
#                key = auth.derive_key(wampPassword,
#                                      challenge.extra['salt'],
#                                      challenge.extra['iterations'],
#                                      challenge.extra['keylen'])
#            else:
#                # plain, unsalted secret
#                key = wampPassword
#
#            # compute signature for challenge, using the key
#            signature = auth.compute_wcs(key, challenge.extra['challenge'])
#
#            # return the signature to the router for verification
#            return signature
#
#        else:
#            raise Exception("Invalid authmethod {}".format(challenge.method))
#
#
#    @inlineCallbacks
#    def onJoin(self, details):
#        out.info('Router session joined')
#
#        yield self.subscribe(self.updatesPending, self.uriPrefix + 'updatesPending')
#        yield self.register(self.update, self.uriPrefix + 'update')
#        yield BaseSession.onJoin(self, details)
#
#
#    def onLeave(self, details):
#        out.info("Router session left: {}".format(details))
#        nexus.core.wamp_connected = False
#        self.disconnect()
#
#
#    def onDisconnect(self):
#        out.info("Router session disconnected.")
#
#    def update(self, pdid, data):
#        print("Sending command {} to pdinstall".format(data['command']))
#        success = pdinstall.sendCommand(data['command'], data)
#
#        # NOTE: If successful, this process will probably be going down soon.
#        if success:
#            return "Sent command to pdinstall"
#        else:
#            return "Sending command to pdinstall failed"
#
#
#    @classmethod
#    def set_update_fetcher(cls, update_fetcher):
#        WampSession.update_fetcher = update_fetcher
#
#
#    @inlineCallbacks
#    def updatesPending(self, pdid):
#        out.info('Notified of updates...')
#        if WampSession.update_fetcher is not None:
#            out.info('Pulling updates from the server...')
#            yield WampSession.update_fetcher.pull_update()



##################################################################
# Copyright 2013-2016 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
    This file contains any settings required by ANY and ALL modules of the paradrop system.
    They are defaulted to some particular value and can be called by any module in the paradrop
    system with the following code:

        from paradrop import settings
        print(settings.STUFF)

    These settings can be overriden by a KYE:VALUE array

    If settings need to be changed, they should be done so by the initialization code
    (such as pdfcd, pdfc_config, etc...)

    This is done by calling the following function:
        settings.updateSettings(settings_array)
"""

import configparser
import os
import sys
import types

from . import constants
from paradrop.lib.utils import pdosq


DEBUG_MODE = False
# VERBOSE = False

CONFIG_HOME_DIR = '/etc/paradrop/'
RUNTIME_HOME_DIR = '/var/run/paradrop/'
TMP_DIR = '/tmp'

#
# paths to store daemon related information
#
LOG_DIR = CONFIG_HOME_DIR + 'logs/'
KEY_DIR = CONFIG_HOME_DIR + 'keys/'
MISC_DIR = CONFIG_HOME_DIR + 'misc/'
CONFIG_FILE = CONFIG_HOME_DIR + 'config'

#
# pdserver
#
PDSERVER = "https://paradrop.org"
PDSERVER_MAX_CONCURRENT_REQUESTS = 2
WAMP_ROUTER = "wss://paradrop.org/ws"

#
# pdfcd
#
PDCONFD_WRITE_DIR = RUNTIME_HOME_DIR + 'pdconfd/'
PDCONFD_ENABLED = True

#
# fc
#
FC_CHUTESTORAGE_FILE = CONFIG_HOME_DIR + "chutes"
FC_CHUTESTORAGE_SAVE_TIMER = 0
FC_BOUNCE_UPDATE = None

DEFAULT_LAN_ADDRESS = "10.0.0.1"
DEFAULT_LAN_NETWORK = "10.0.0.0/24"
DYNAMIC_NETWORK_POOL = "10.128.0.0/9"

#
# uci
#
# These should be in a persistent directory (not /var/run, /tmp, etc.) because
# we expect to find them at startup.
#
UCI_CONFIG_DIR = CONFIG_HOME_DIR +  "uci/config/"
UCI_BACKUP_DIR = CONFIG_HOME_DIR + "uci/config-backup.d/"

#
# local portal
#
PORTAL_SERVER_PORT = 8080

#
# Local domain - this domain and subdomains will be resolved to the router so
# that chutes and their users can access it by name.
#
LOCAL_DOMAIN = "paradrop.io"

#
# hostconfig
#
HOST_CONFIG_FILE = "/etc/paradrop/hostconfig.yaml"
DEFAULT_HOST_CONFIG_FILE = "/etc/paradrop/hostconfig.default.yaml"
HOST_DATA_PARTITION = "/"

#
# Chute data directory is used to provide persistence for chute data.
# The developer can store files here and they will persist across chute
# updates and restarts (but not uninstallation).
#
# Note: we do not want this in SNAP_DATA because that path
# contains the snap version.  We want this path to stay constant across
# paradrop upgrades because chutes will have volumes mounted here.
#
# Internal is inside the chute; external is in the host.
#
INTERNAL_DATA_DIR = "/data/"
EXTERNAL_DATA_DIR = "~/.paradrop/chute-data/{chute}/"

#
# System directory is used to share system information from the host
# down to the chute such as a list of devices connected to WiFi.  This
# is mounted read-only inside the chute.
#
# Internal is inside the chute; external is in the host.
#
INTERNAL_SYSTEM_DIR = "/paradrop/"
EXTERNAL_SYSTEM_DIR = "/var/run/paradrop/system/{chute}/"
# TODO: for local mode?

#
# Username and password to access the private registry.
#
# TODO: The router should receive credentials from pdserver or use something
# unique like its API key.
#
REGISTRY_USERNAME = "routers"
REGISTRY_PASSWORD = "zai7geigh0ujahQu"

# Reject by default chute updates that would install an older version.  This
# has been set to False to fulfill the requirement that users be able to
# downgrade chutes.
REJECT_DOWNGRADE = False

DOCKER_BIN_DIR = "/usr/bin"

# Interface (e.g. Unix socket) to use to access snapd API.
SNAPD_INTERFACE = "/run/snapd.socket"

# Filename to search for in chute project directory for configuring how we
# build the chute.
CHUTE_CONFIG_FILE = "paradrop.yaml"

# UID for unprivileged containers.  File ownership on mounted volumes needs to
# be consistent between the container and host.
CONTAINER_UID = 999

# Directory containing Airshark installation ('airshark' binary and
# 'run_airshark.sh' script).
AIRSHARK_INSTALL_DIR = "/snap/airshark/current"

# Boolean flag to enable/disable concurrent builds for Docker images.  If
# enabled, the update pipeline will yield during a build to allow another
# update to make progress. This should improve the experience for multi-user
# access, but the interleaving of updates could cause subtle issues.
CONCURRENT_BUILDS = True

# Boolean flag to enable/disable monitor mode interfaces for chutes. This is by
# default disabled because monitor mode interfaces are dangerous.  They enable
# malicious chutes to record network traffic, and furthermore, the feature
# itself is experimental. There may be issues with kernel drivers or our
# implementation that cause system instability.
ALLOW_MONITOR_MODE = False

# Check if Docker daemon is in a bad state (process not running but pid file
# exists).  This does not work in strict confinement.
CHECK_DOCKER = False

# Directory for ZeroTier runtime files. We can find the API authtoken in a file
# in this directory.
ZEROTIER_LIB_DIR = "/var/lib/zerotier-one"

# Default username and password for the admin panel.
DEFAULT_PANEL_USERNAME = "paradrop"
DEFAULT_PANEL_PASSWORD = ""

# Default wireless settings used for hostconfig generation.  If
# DEFAULT_WIRELESS_ENABLED is set to False, we will not create an access point
# by default.
DEFAULT_WIRELESS_ENABLED = True
DEFAULT_WIRELESS_ESSID = "Paradrop"
DEFAULT_WIRELESS_KEY = None

GOVERNOR_INTERFACE = "/var/run/governor.socket"

###############################################################################
# Helper functions
###############################################################################


def iterate_module_attributes(module):
    """
    Iterate over the attributes in a module.

    This is a generator function.

    Returns (name, value) tuples.
    """
    for name in dir(module):
        # Ignore fields marked as private or builtin.
        if name.startswith('_'):
            continue

        value = getattr(module, name)

        # Ignore callable objects (functions) and loaded modules.
        if callable(value) or isinstance(value, types.ModuleType):
            continue

        yield (name, value)


def load_from_file(path):
    """
    Load settings from an INI file.

    This will check the configuration file for a lowercase version of all of
    the settings in this module. It will look in a section called "base".

    The example below will set PORTAL_SERVER_PORT.

        [base]
        portal_server_port = 4444
    """
    config = configparser.SafeConfigParser()
    config.read(path)

    mod = sys.modules[__name__]
    for name, _ in iterate_module_attributes(mod):
        # Check if lowercase version exists in the file and load the
        # appropriately-typed value.
        key = name.lower()
        if config.has_option(constants.BASE_SETTINGS_SECTION, key):
            value = config.get(constants.BASE_SETTINGS_SECTION, key)
            setattr(mod, name, parseValue(value))


def parseValue(key):
    """
    Attempts to parse the key value, so if the string is 'False' it will parse a boolean false.

    :param key: the key to parse
    :type key: string

    :returns: the parsed key.
    """
    # Is it a boolean?
    if(key == 'True' or key == 'true'):
        return True
    if(key == 'False' or key == 'false'):
        return False

    # Is it None?
    if(key == 'None' or key == 'none'):
        return None

    # Is it a float?
    if('.' in key):
        try:
            f = float(key)
            return f
        except:
            pass

    # Is it an int?
    try:
        i = int(key)
        return i
    except:
        pass

    # TODO: check if json

    # Otherwise, its just a string:
    return key


def updatePaths(configHomeDir, runtimeHomeDir="/var/run/paradrop"):
    # Get a handle to our settings defined above
    mod = sys.modules[__name__]

    mod.CONFIG_HOME_DIR = configHomeDir
    mod.RUNTIME_HOME_DIR = runtimeHomeDir
    mod.FC_CHUTESTORAGE_FILE = os.path.join(mod.CONFIG_HOME_DIR, "chutes")
    mod.EXTERNAL_DATA_DIR = os.path.join(mod.CONFIG_HOME_DIR, "chute-data/{chute}/")
    mod.EXTERNAL_SYSTEM_DIR = os.path.join(runtimeHomeDir, "system", "{chute}")
    mod.LOG_DIR = os.path.join(mod.CONFIG_HOME_DIR, "logs/")
    mod.KEY_DIR = os.path.join(mod.CONFIG_HOME_DIR, "keys/")
    mod.MISC_DIR = os.path.join(mod.CONFIG_HOME_DIR, "misc/")
    mod.CONFIG_FILE = os.path.join(mod.CONFIG_HOME_DIR, "config")
    mod.HOST_CONFIG_FILE = os.path.join(mod.CONFIG_HOME_DIR, "hostconfig.yaml")
    mod.DEFAULT_HOST_CONFIG_FILE = os.path.join(mod.CONFIG_HOME_DIR, "hostconfig.default.yaml")
    mod.UCI_CONFIG_DIR = os.path.join(mod.CONFIG_HOME_DIR, "uci/config.d/")
    mod.UCI_BACKUP_DIR = os.path.join(mod.CONFIG_HOME_DIR, "uci/config-backup.d/")
    mod.PDCONFD_WRITE_DIR = os.path.join(mod.RUNTIME_HOME_DIR, 'pdconfd')


def loadSettings(mode="local", slist=[]):
    """
    Take a list of key:value pairs, and replace any setting defined.
    Also search through the settings module and see if any matching
    environment variables exist to replace as well.

    :param slist: the list of key:val settings
    :type slist: array.

    :returns: None
    """

    # Get a handle to our settings defined above
    mod = sys.modules[__name__]

    # Adjust default paths if we are running under ubuntu snappy
    snapPath = os.environ.get("SNAP", None)
    snapCommonPath = os.environ.get("SNAP_COMMON", None)
    snapDataPath = os.environ.get("SNAP_DATA", None)

    # Directories where we might find a settings.ini file.
    settings_file_dirs = [".", "/etc"]

    if mode == "local":
        updatePaths(os.path.join(os.path.expanduser("~"), ".paradrop/"),
                                 "/tmp/.paradrop/")
        mod.HOST_DATA_PARTITION = mod.CONFIG_HOME_DIR
    elif mode == "unittest":
        updatePaths("/tmp/.paradrop-test/", "/tmp/.paradrop-test/")
        mod.HOST_DATA_PARTITION = mod.CONFIG_HOME_DIR
    elif snapCommonPath is not None:
        settings_file_dirs.append(snapCommonPath)
        updatePaths(snapCommonPath, snapDataPath)
        mod.HOST_DATA_PARTITION = "/writable"
        mod.DOCKER_BIN_DIR = "/snap/bin"
        mod.GOVERNOR_INTERFACE = os.path.join(snapPath, "governor", "governor.socket")
        mod.ZEROTIER_LIB_DIR = os.path.join(snapPath, "zerotier-one")

    for x in [mod.LOG_DIR, mod.KEY_DIR, mod.MISC_DIR]:
        pdosq.makedirs(x)

    # First overwrite settings they may have provided with the arg list
    for kv in slist:
        k, v = kv.split(':', 1)
        # We can either replace an existing setting, or set a new value, we don't care
        setattr(mod, k, parseValue(v))

    # Next check for settings from file(s) which may be located in a few
    # different directories.
    for d in settings_file_dirs:
        path = os.path.join(d, constants.SETTINGS_FILE_NAME)
        load_from_file(path)

    # Now search through our settings and look for environment variable matches
    # they defined. Environment variables override all other sources.
    for name, _ in iterate_module_attributes(mod):
        # Check for an environment variable matching the setting.
        value = os.environ.get(name, None)
        if value is not None:
            setattr(mod, name, parseValue(value))



'''
Exceptions and their subclasses

TODO: Distill these down and make a heirarchy.
'''


class PdServerException(Exception):
    pass


class InteralException(Exception):
    pass


class AuthenticationError(PdServerException):
    pass


class InvalidCredentials(PdServerException):
    pass


class PdidError(PdServerException):
    pass


class PdidExclusionError(PdServerException):
    pass


class ModelNotFound(PdServerException):
    pass


class ParadropException(Exception):
    pass


class ChuteNotFound(ParadropException):
    pass


class ServiceNotFound(ParadropException):
    def __init__(self, name):
        message = "Service ({}) not found".format(name)
        super(ServiceNotFound, self).__init__(message)


class ChuteNotRunning(ParadropException):
    pass


class DeviceNotFoundException(ParadropException):
    pass



###################################################################
# Copyright 2013-2016 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
This module generates update plans for router operations such as factory reset.
"""

from paradrop.base.output import out
from paradrop.core.config import power, state
from paradrop.core.container import dockerapi

from . import plangraph


def generatePlans(update):
    out.verbose("%r\n" % (update))

    if update.updateType == "factoryreset":
        update.plans.addPlans(plangraph.STATE_CALL_STOP,
                              (dockerapi.removeAllContainers, ))
        update.plans.addPlans(plangraph.STATE_SAVE_CHUTE,
                              (state.removeAllChutes, ))

    elif update.updateType == "reboot":
        update.plans.addPlans(plangraph.STATE_CALL_STOP,
                              (power.reboot, ))

    elif update.updateType == "shutdown":
        update.plans.addPlans(plangraph.STATE_CALL_STOP,
                              (power.shutdown, ))



###################################################################
# Copyright 2013-2017 All Rights Reserved
# Authors: The ParaDrop Team
###################################################################

"""
Output mapping, capture, storange, and display.

Some of the methods and choice here may seem strange -- they are meant to
keep this file in
"""

import colorama
import json
import os
import queue
import sys
import threading
import time
import traceback

import six
import smokesignal

from enum import Enum
from twisted.python.logfile import DailyLogFile
from twisted.python import log

from . import pdutils


# colorama package does colors but doesn't do style, so keeping this for now
BOLD = '\033[1m'
LOG_NAME = 'log'

Level = Enum('Level', 'HEADER, VERBOSE, INFO, PERF, WARN, ERR, SECURITY, FATAL, USAGE')

# Represents formatting information for the specified log type
LOG_TYPES = {
    Level.HEADER: {'name': Level.HEADER.value, 'glyph': '==', 'color': colorama.Fore.BLUE},
    Level.VERBOSE: {'name': Level.VERBOSE.value, 'glyph': '>>', 'color': colorama.Fore.BLACK},
    Level.INFO: {'name': Level.INFO.value, 'glyph': '--', 'color': colorama.Fore.GREEN},
    Level.PERF: {'name': Level.PERF.value, 'glyph': '--', 'color': colorama.Fore.WHITE},
    Level.WARN: {'name': Level.WARN.value, 'glyph': '**', 'color': colorama.Fore.YELLOW},
    Level.ERR: {'name': Level.ERR.value, 'glyph': '!!', 'color': colorama.Fore.RED},
    Level.SECURITY: {'name': Level.SECURITY.value, 'glyph': '!!', 'color': BOLD + colorama.Fore.RED},
    Level.FATAL: {'name': Level.FATAL.value, 'glyph': '!!', 'color': colorama.Back.WHITE + colorama.Fore.RED},
    Level.USAGE: {'name': Level.USAGE.value, 'glyph': '++', 'color': colorama.Fore.CYAN},
}

###############################################################################
# Logging Utilities
###############################################################################


def silentLogPrefix(stepsUp):
    '''
    logPrefix v2-- gets caller information silently (without caller intervention)
    The single parameter reflects how far up the stack to go to find the caller and
    depends how deep the direct caller to this method is wrt to the target caller

    NOTE: Some calls cannot be silently prefixed (getting into the twisted code is a
    great example)

    :param stepsUp: the number of steps to move up the stack for the caller
    :type steps: int.
    '''

    try:
        trace = sys._getframe(stepsUp).f_code.co_filename
        line = sys._getframe(stepsUp).f_lineno
        module, package = parseLogPrefix(trace)
    except:
        return 'unknown', 'unknown', '??'

    return package, module, line


def parseLogPrefix(tb):
    '''
    Takes a traceback returned by 'extract_tb' and returns the package, module,
    and line number
    '''
    path = tb.split('/')
    module = path[-1].replace('.py', '')
    package = path[-2]

    return module, package


class PrintLogThread(threading.Thread):

    '''
    All file printing access from one thread.

    Receives information when its placed on the passed queue.
    Called from one location: Output.handlePrint.

    Does not close the file: this happens in Output.endLogging. This
    simplifies the operation of this class, since it only has to concern
    itself with the queue.

    The path must exist before DailyLog runs for the first time.
    '''

    def __init__(self, path, queue, name):
        threading.Thread.__init__(self)
        self.queue = queue
        self.writer = DailyLogFile(name, path)

        # Don't want this to float around if the rest of the system goes down
        self.setDaemon(True)

    def run(self):
        while True:
            result = self.queue.get(block=True)

            try:
                writable = json.dumps(result)
                self.writer.write(writable + '\n')
                self.writer.flush()
            except:
                pass

            self.queue.task_done()


class OutputRedirect(object):

    """
    Intercepts passed output object (either stdout and stderr), calling the provided callback
    method when input appears.

    Retains the original mappings so writing can still happen. Performs no formatting.
    """

    def __init__(self, output, contentAppearedCallback, logType):
        self.callback = contentAppearedCallback
        self.trueOut = output
        self.type = logType

    def trueWrite(self, contents):
        ''' Someone really does want to output'''
        formatted = str(contents)

        # print statement ususally handles these things
        if len(formatted) == 0 or formatted[-1] is not '\n':
            formatted += '\n'

        self.trueOut.write(formatted)

    def flush(self):
        self.trueOut.flush()

    def write(self, contents):
        '''
        Intercept output to the assigned target and callback with it. The true output is
        returned with the callback so the delegate can differentiate between captured outputs
        in the case when two redirecters are active.
        '''
        if contents == '\n':
            return

        package, module, line = silentLogPrefix(2)

        ret = { \
            'message': str(contents.strip()), \
            'type': self.type['name'], \
            'extra': {'details': 'floating print statement'}, \
            'package': package, \
            'module': module, \
            'timestamp': time.time(), \
            'pdid': 'UNSET', \
            'line': line \
        }

        self.callback(ret)


###############################################################################
# Output Classes
###############################################################################

class BaseOutput(object):

    '''
    Base output type class.

    This class and its subclasses are registered with an attribute on the global
    'out' function and is responsible for formatting the given output stream
    and returning it as a "log structure" (which is a dict.)

    For example:
        out.info("Text", anObject)

    requires a custom object to figure out what to do with anObject where the default case will simply
    parse the string with an appropriate color.

    Objects are required to output a dict that mininmally contains the keys message and type.
    '''

    def __init__(self, logType):
        '''
        Initialize this output type.

        :param logType: how this output type is displayed
        :type logType: dictionary object containing name, glyph, and color keys
        '''

        self.type = logType

    def __call__(self, args, logPrefixLevel=3, **extra):
        '''
        Called as an attribute on out. This method takes the passed params and builds a log dict,
        returning it.

        Subclasses can customize args to include whatever they'd like, adding content
        under the key 'extras.' The remaining keys should stay in place.
        '''
        package, module, line = silentLogPrefix(3)

        # String newlines
        if args[-1] == '\n':
            args = args.strip()

        ret = { \
            'message': str(args), \
            'type': self.type['name'], \
            'extra': extra, \
            'package': package, \
            'module': module, \
            'timestamp': time.time(), \
            'pdid': 'UNSET', \
            'line': line \
        }

        return ret

    def formatOutput(self, logDict):
        '''
        Convert a logdict into a custom formatted, human readable version suitable for
        printing to console.
        '''
        trace = '[%s.%s#%s @ %s] ' % ( \
            logDict['package'], \
            logDict['module'], \
            logDict['line'], \
            pdutils.stimestr(logDict['timestamp']) \
        )
        return self.type['color'] + 'PARADROP ' + self.type['glyph'] + ' ' \
            + trace + logDict['message'] + colorama.Style.RESET_ALL

    def __repr__(self):
        return "REPR"


class TwistedOutput(BaseOutput):

    # There's a host of things we simply don't care about. This is bad form
    # and not great for performance. Alternatives welcome.
    blacklist = [
        'Starting factory',
        'Stopping factory',
        'Log opened'
    ]

    def __call__(self, args):
        '''
        Catch twisted logs and make them fall in line with our logs.

        Ignore exceptions (those get their own handler)

        Twisted will always pass a dict and guarantees [message, isError, and printed]
        will be in there.
        '''

        # there is another class responsible for handling error messages. Ignore these.
        if args['isError'] == 1:
            return None

        # Start with the default message, but just grab the whole thing if it doesn't work
        try:
            message = args['message'][0]
        except:
            # For now, lets ignore the big messages (ones without a well-defined message,
            # generally these are internal twisted messages we may not care about)
            return None
            message = str(args)

        for x in TwistedOutput.blacklist:
            if x in message:
                return None

        ret = { \
            'message': message, \
            'type': self.type['name'], \
            'extra': {}, \
            'package': 'twisted', \
            'module': 'internal', \
            'timestamp': time.time(), \
            'pdid': 'UNSET', \
            'line': '??' \
        }

        return ret


class TwistedException(BaseOutput):

    def __call__(self, args):
        '''
        Catch twisted logs and make them fall inline with our logs.

        Only catch errors.

        Twisted will always pass a dict and guarantees [message, isError, and printed]
        will be in there.
        '''

        if args['isError'] == 0:
            return None

        # special handling of exception prefix logging
        try:
            tb = args['failure'].getTracebackObject()
            stacktrace = traceback.extract_tb(tb)[-1]

            module, package = parseLogPrefix(stacktrace[0])
            line = stacktrace[1]
        except Exception:
            package, module, line = "uknown", 'unknown', '??'

        ret = { \
            'message': str(args['failure'].getTraceback().strip()), \
            'type': self.type['name'], \
            'extra': {'details': 'floating print statement'}, \
            'package': package, \
            'module': module, \
            'timestamp': time.time(), \
            'pdid': 'UNSET', \
            'line': line \
        }

        return ret


class ExceptionOutput(BaseOutput):

    '''
    Handle vanilla exceptions passed directly to us using out.exception
    '''

    def __call__(self, exception, random):
        '''
        The variable 'Random' is a leftover from the previous implementation and should be removed.
        '''

        # print exception.__traceback_
        ex_type, ex, tb = sys.exc_info()
        trace = traceback.extract_tb(tb)
        lastFrame = trace[-1]

        package, module = parseLogPrefix(lastFrame[0])
        line = lastFrame[1]

        message = type(exception).__name__ + ': ' + str(exception) + '\n'

        for x in trace:
            message += '  File "%s", line %d, in %s\n\t%s\n' % (x[0], x[1], x[2], x[3])

        ret = { \
            'message': message, \
            'type': self.type['name'], \
            'extra': {'details': 'floating print statement'}, \
            'package': package, \
            'module': module, \
            'timestamp': time.time(), \
            'pdid': 'UNSET', \
            'line': line \
        }

        return ret


class Output(object):

    '''
    Class that allows stdout/stderr trickery.
    By default the paradrop object will contain an @out variable
    (defined below) and it will contain 2 members of "err" and "fatal".

    Each attribute of this class should be a function which points
    to a class that inherits IOutput(). We call these functions
    "output streams".

    The way this Output class is setup is that you pass it a series
    of kwargs like (stuff=OutputClass()). Then at any point in your
    program you can call "paradrop.out.stuff('This is a string\n')".

    This way we can easily support different levels of verbosity without
    the need to use some kind of bitmask or anything else. On-the-fly output
    creation is no longer supported due to the metadata and special processing
    added. It is still possible, but not implemented.

    This is done by the __getattr__ function below, basically in __init__ we set
    any attributes you pass as args, and anything else not defined gets sent to __getattr__
    so that it doesn't error out.
    '''

    def __init__(self, **kwargs):
        """Setup the initial set of output stream functions."""

        # Begins intercepting output and converting ANSI characters to win32 as applicable
        colorama.init()

        # Refactor this as an Output class
        self.__dict__['redirectErr'] = OutputRedirect(sys.stderr, self.handlePrint, LOG_TYPES[Level.VERBOSE])
        self.__dict__['redirectOut'] = OutputRedirect(sys.stdout, self.handlePrint, LOG_TYPES[Level.VERBOSE])

        # by default, dont steal output and print to console
        self.stealStdio(False)
        self.logToConsole(True)

        # Setattr wraps the output objects in a
        # decorator that allows this class to intercept their output, This dict holds the
        # original objects.
        self.__dict__['outputMappings'] = {}

        for name, func in six.iteritems(kwargs):
            setattr(self, name, func)

    def __getattr__(self, name):
        """Catch attribute access attempts that were not defined in __init__
            by default throw them out."""

        # raise NotImplementedError("You must create " + name + " to log with it")
        pass

    def __setattr__(self, name, val):
        def inner(*args, **kwargs):
            result = val(*args, **kwargs)
            self.handlePrint(result)
            return result

        # can't call setattr here (which normally looks like self.name = inner)
        self.__dict__[name] = inner

        # Save the original function (unwrapped) under the tag its registered with
        # so we can later query the objects by this tag and ask them to print
        self.__dict__['outputMappings'][name] = val

    def __repr__(self):
        return "REPR"

    def startLogging(self, filePath=None, stealStdio=False, printToConsole=True):
        '''
        Begin logging. The output class is ready to go out of the box, but in order
        to prevent mere imports from stealing stdio or console logging to vanish
        these must be manually turned on.

        :param filePath: if provided, begin logging to the given directory. If
            not provided, do not write out logs.
        :type filePath: str
        :param stealStdio: choose to intercept stdio (including vanilla print
            statements) or allow it to passthrough
        :type stealStdio: bool.
        :param printToConsole: output the results of all logging to the console. This
            is primarily a performance consideration when running in production
        :type printToConsole: bool.

        '''

        # Initialize printer thread
        self.__dict__['logpath'] = None

        if filePath is not None:
            self.__dict__['queue'] = queue.Queue()
            self.__dict__['printer'] = PrintLogThread(filePath, self.queue, LOG_NAME)
            self.__dict__['logpath'] = filePath
            self.printer.start()

        # by default, stdio gets captures. This can be toggled off
        self.stealStdio(stealStdio)
        self.logToConsole(printToConsole)

        # Override twisted logging (allows us to cleanly catch all exceptions)
        # This must come after the setattr calls so we get the wrapped object
        log.startLoggingWithObserver(self.twisted, setStdout=False)
        log.startLoggingWithObserver(self.twistedErr, setStdout=False)

    def endLogging(self):
        '''
        Ask the printing thread to flush and end, then return.
        '''

        out.info('Asking file logger to close')
        self.queue.join()

        # Because the print thread can't tell when it goes down as currently designed
        self.printer.writer.close()

    def handlePrint(self, logDict):
        '''
        All printing objects return their messages. These messages are routed
        to this method for handling.

        Send the messages to the printer. Optionally display the messages.
        Decorate the print messages with metadata.

        :param logDict: a dictionary representing this log item. Must contain keys
        message and type.
        :type logDict: dict.
        '''

        # If the logger returns None, assume we dont want the output
        if logDict is None:
            return

        # write out the log message to file
        if self.queue is not None:
            self.queue.put(logDict)

        res = self.messageToString(logDict)

        # Write out the human-readable version to out if needed (but always print out
        # exceptions for testing purposes)
        if self.printLogs or logDict['type'] == 'ERR':
            self.redirectOut.trueWrite(res)

        # Broadcast the log to interested parties
        smokesignal.emit('logs', logDict)

    def messageToString(self, message):
        '''
        Converts message dicts to a format suitable for printing based on
        the conversion rules laid out in in that class's implementation.

        :param message: the dict to convert to string
        :type message: dict.
        :returns: str
        '''

        level = Level(message['type'])
        outputObject = self.outputMappings[level.name.lower()]
        return outputObject.formatOutput(message)

    def getLogsSince(self, target, purge=False):
        '''
        Reads all logs and returns their contents. The current log file is not touched.
        Removes old log files if 'purge' is set (though this is a topic for debate...)

        The server will be most interested in this call, but it needs to register for
        new logs first, else there's a good chance to see duplicates.

        NOTE: don't open all log files, check to open only the ones that might be relevant.
        This is certainly a bug and can cause memory issues.

        :param target: seconds since the GMT epoch. Method returns logs that have timestamps later than this.
        :type target: float.
        :param purge: deletes the old log files (except today's) if set
        :type purge: bool.
        :returns: a list of dictionaries containing log information. Not ordered.
        '''

        if not self.logpath:
            out.warn('Asked for log files, but this instance of the output class '
                     'is not currently configured for file logging. '
                     'Call startLogging with a directory first! ')
            return

        ret = []

        for f in os.listdir(self.logpath):
            path = self.logpath + '/' + f

            # the current log file is treated differently (no date, no delete)
            if f != LOG_NAME:
                t = time.strptime(f.split('.')[1], '%Y_%m_%d')

                # dont load those with times earlier than target
                if t >= target:
                    with open(path, 'r') as x:
                        ret += [json.loads(y) for y in x.readlines()]

                # delete all files except log once read
                if purge:
                    os.remove(path)

            # current log file is always loaded
            else:
                with open(path, 'r') as f:
                    ret += [json.loads(y) for y in f.readlines()]

        ret = filter(lambda x: x['timestamp'] > target, ret)

        return ret


    ###############################################################################
    # Reconfiguration
    ###############################################################################

    def stealStdio(self, newStatus):
        self.__dict__['stealIo'] = newStatus

        if newStatus:
            # assign our interceptor objects to the outputs
            sys.stdout = self.redirectOut
            sys.stderr = self.redirectErr
        else:
            # return stdout and err to their respective positions
            sys.stdout = self.__dict__['redirectOut'].trueOut
            sys.stderr = self.__dict__['redirectErr'].trueOut

    def logToConsole(self, newStatus):
        self.__dict__['printLogs'] = newStatus


out = Output(
    header=BaseOutput(LOG_TYPES[Level.HEADER]),
    testing=BaseOutput(LOG_TYPES[Level.VERBOSE]),
    verbose=BaseOutput(LOG_TYPES[Level.VERBOSE]),
    info=BaseOutput(LOG_TYPES[Level.INFO]),
    usage=BaseOutput(LOG_TYPES[Level.USAGE]),
    perf=BaseOutput(LOG_TYPES[Level.PERF]),
    warn=BaseOutput(LOG_TYPES[Level.WARN]),
    err=BaseOutput(LOG_TYPES[Level.ERR]),
    exception=ExceptionOutput(LOG_TYPES[Level.ERR]),
    security=BaseOutput(LOG_TYPES[Level.SECURITY]),
    fatal=BaseOutput(LOG_TYPES[Level.FATAL]),
    twisted=TwistedOutput(LOG_TYPES[Level.INFO]),
    twistedErr=TwistedException(LOG_TYPES[Level.ERR])
)



"""
Quiet pdos module.
Implements utility OS operations without relying on the output module.
Therefore, this module can be used by output without circular dependency.
"""
from __future__ import print_function

import errno
import os

from paradrop.lib.utils.yaml import yaml


def makedirs(p):
    """
    Recursive directory creation (like mkdir -p).
    Returns True if the path is successfully created, False if it existed
    already, and raises an OSError on other error conditions.
    """
    try:
        os.makedirs(p)
        return True
    except OSError as e:
        # EEXIST is fine (directory already existed).  We also occasionally get
        # EROFS when making directories in SNAP_COMMON. I am not quite sure
        # what is going on there. Re-raise any other errors.
        if e.errno == errno.EEXIST:
            pass
        elif e.errno == errno.EROFS:
            print("Read-only filesystem error occurred while making {}".format(p))
        else:
            raise e
    return False


def read_yaml_file(path, default=None):
    """
    Read the contents of a file and interpret as YAML.

    default: return value if the file cannot be read.
    """
    try:
        with open(path, "r") as source:
            return yaml.load(source)
    except IOError:
        return default
    except OSError:
        return default


def safe_remove(path):
    """
    Remove a file or silently pass if the file does not exist.

    This function has the same effect as os.remove but suppresses the error if
    the file did not exist. Notably, it must not be used to remove directories.

    Returns True if a file was removed or False if no file was removed.
    """
    try:
        os.remove(path)
        return True
    except OSError as err:
        # Suppress the exception if it is a file not found error.
        # Otherwise, re-raise the exception.
        if err.errno != errno.ENOENT:
            raise

    return False



"""
This file contains constants that may be used safely throughout the
code.  Although Python does not have any native mechanisms for enforcing
read-only variables, a violation of this convention would be easy enough
to spot.

    # Please do not do this.
    from paradrop.base import constants
    constants.IMPORTANT_CONSTANT = "boo"

Some of these variables were moved from paradrop.base.settings in an
effort to clean up that module when it became clear that there were two
types of variables in that module, variables that could reasonably be
considered settings for a user or administrator, and variables that were
actually intended to be constants.

When considering whether to put a new variable here or in the settings
module, consider the following factors:

    1. If a user were to change this setting on a running system, would
    it result in unexpected or undefined behavior?

    2. Should changing the value of this variable be subject to a
    version-controlled code change.

    3. Would an advanced user or administrator have a use case for
    changing this variable?

As an example of something that needs to be a constant, consider
the reserved chute name, RESERVED_CHUTE_NAME, which is used to mark
settings in UCI files as belonging to the system rather than a chute.
If someone were to change this value on a running system, the resulting
behavior would be undefined, so this value belongs in constants rather
than settings.
"""

# Name of section in the settings file for paradrop.base.settings values.
BASE_SETTINGS_SECTION = "base"

# Space-separated list of features that will be passed to chutes through
# environment variable or API call. This allows chutes to check the features of
# the host on which they are running. Update this when significant features are
# added.
DAEMON_FEATURES = "audio hostapd-control"

# Character limit for network interface names. The limit is imposed by the
# operating system.
MAX_INTERFACE_NAME_LEN = 15

# This is not actually a chute, but a reserved name that is used to mark
# settings and changes that belong to the system rather than a chute.
RESERVED_CHUTE_NAME = "__PARADROP__"

# Name of settings file used by paradrop.base.settings.
SETTINGS_FILE_NAME = "settings.ini"



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import heapq

###############################################################################
# PRIORITYFLAGS: Define the priority numbers as constants here
#
# These priority numbers control the order of execution for the stages of an
# update.  Many of them are unused and can be repurposed.  Feel free to insert
# new steps as needed.  However, be very careful changing the order of existing
# steps, as many of them require a particular ordering.
###############################################################################

###############################################################################
# Security Checks
ENFORCE_ACCESS_RIGHTS           = 1

###############################################################################
# Configuration Generation
DOWNLOAD_CHUTE_FILES            = 6
LOAD_CHUTE_CONFIGURATION        = 7
STRUCT_GET_SYSTEM_DEVICES       = 8
STRUCT_GET_RESERVATIONS         = 9
STRUCT_GET_HOST_CONFIG          = 10
STRUCT_GET_INT_NETWORK          = 11
STRUCT_GET_OS_NETWORK           = 12
STRUCT_GET_OS_WIRELESS          = 13
STRUCT_GET_L3BRIDGE_CONFIG      = 14
TRAFFIC_GET_OS_FIREWALL         = 18
TRAFFIC_GET_DEVELOPER_FIREWALL  = 19
RUNTIME_GET_VIRT_PREAMBLE       = 20
RUNTIME_GET_VIRT_DHCP           = 21
DHCP_GET_VIRT_RULES             = 22
STATE_GENERATE_SERVICE_PLANS    = 23

###############################################################################
# Build Docker Image
STATE_BUILD_IMAGE               = 30
STATE_CHECK_IMAGE               = 31
STATE_CALL_STOP                 = 35

###############################################################################
# System Validation
CHECK_SYSTEM_DEVICES            = 40
RESOURCE_GET_ALLOCATION         = 41

###############################################################################
# Write Configuration Data
RUNTIME_RELOAD_CONFIG_BACKOUT   = 45
CREATE_VOLUME_DIRS              = 50
STRUCT_SET_SYSTEM_DEVICES       = 51
STRUCT_SET_HOST_CONFIG          = 52
STRUCT_SET_OS_WIRELESS          = 53
STRUCT_SET_OS_NETWORK           = 54
TRAFFIC_SET_OS_FIREWALL         = 55
STRUCT_SET_L3BRIDGE_CONFIG      = 56
RESOURCE_SET_VIRT_QOS           = 57
NAME_MNT_PART                   = 58
STATE_SET_VIRT_CONFIG           = 59
STATE_SET_VIRT_SCRIPT           = 60
RUNTIME_SET_VIRT_DHCP           = 61
DHCP_SET_VIRT_RULES             = 62
RUNTIME_RELOAD_CONFIG           = 63

###############################################################################
# Operations On Configuration Changes
STATE_CALL_CFG                  = 70
FILES_COPY_FROM_OS              = 71
ZEROTIER_CONFIGURE              = 72
AIRSHARK_CONFIGURE              = 73
TELEMETRY_SERVICE               = 74
APPLY_WAIT_ONLINE_FIX           = 75
FILES_COPY_USER                 = 76
FILES_COPY_TO_MNT               = 77
STARTUP_CALL_EARLY_START        = 78
STRUCT_RELOAD_NETWORK           = 79
STRUCT_RELOAD_WIFI              = 80
TRAFFIC_RELOAD_FIREWALL         = 81
DHCP_RELOAD                     = 82
STATE_MAKE_EXEC                 = 83

###############################################################################
# Manipulate Chute Container
STATE_NET_STOP                  = 84
STATE_CREATE_BRIDGE             = 85
STATE_CALL_STOP                 = 86
STATE_CALL_START                = 87
STATE_CALL_FREEZE               = 88
STATE_CALL_CLEANUP              = 89
STATE_FILES_START               = 90
STATE_NET_START                 = 91
STATE_REMOVE_BRIDGE             = 92

STATE_SAVE_CHUTE                = 95
RESOURCE_SET_ALLOCATION         = 96

RECONFIGURE_PROXY               = 98
SNAP_INSTALL                    = 99
COAP_CHANGE_PROCESSES           = 100


class Plan:
    """
        Helper class to hold onto the actual plan data associated with each plan
    """
    def __init__(self, func, *args):
        """
            Takes a tuple of (function, (args)) of actions to perform at some point.
        """
        self.func = func
        self.args = args

    def __repr__(self):
        return "<%r>" % (self.func)

    def __eq__(self, o):
        """Implementing equal function so we can do 'if a in b' logic"""
        if(not self.func is o.func):
            return False
        if(self.args != o.args):
            return False
        return True


class PlanMap:
    """
        This class helps build a dependency graph required to determine what steps are
        required to update a Chute from a previous version of its configuration.
    """
    def __init__(self, name):
        """Hold onto a name object so we know what we are referencing with this PlanMap."""
        self.name = name

        # Hold onto all available plans here.
        self.plans = []
        self.abortPlans = []

        # workingPlans is implemented as a heapq. All of the other plan lists
        # are plain lists.
        self.maxPriorityReturned = 0
        self.startedAbort = False
        self.workingPlans = []
        self.workingAbortPlans = []

        self.skipFunctions = []

    def addPlans(self, priority, todoPlan, abortPlan=[]):
        """
            Adds new Plan objects into the list of plans for this PlanMap.

            Arguments:
                @priority   : The priority number (1 is done first, 99 done last - see PRIORITYFLAGS section at top of this file)
                @todoPlan   : A tuple of (function, (args)), this is the function that completes the actual task requested
                              the args can either be a single variable, a tuple of variables, or None.
                @abortPlan  : A tuple of (function, (args)) or a list of tuple or None.
                              This is what should be called if a plan somewhere in the chain
                              fails and we need to undo the work we did here - this function is only called if a higher priority
                              function fails (ie we were called, then something later on fails that would cause us to undo everything
                              we did to setup/change the Chute).
        """
        # It is OK to call addPlans out of order during the plan generation
        # phase, but if we have already started execution and add something out
        # of order, it indicates a programming error.
        if priority < self.maxPriorityReturned:
            raise Exception("Added plan out of order during execution.")

        # They can provide multiple abortPlans in a list, or 1 in a tuple, or none at all empty list
        # But internally the abortPlan should be either None or a list (even if the list is size 1)
        if(isinstance(abortPlan, tuple)):
            abortP = [Plan(*abortPlan)]
        elif(isinstance(abortPlan, list)):
            # See if there is no plans at all
            if(len(abortPlan) == 0):
                abortP = None
            else:
                abortP = [Plan(*a) for a in abortPlan]
        else:
            raise Exception('BadAbortPlan', 'Plan should be tuple, list of tuples, or empty')

        todoP = Plan(*todoPlan)

        # Now add into the set
        self.plans.append((priority, todoP, abortP))
        heapq.heappush(self.workingPlans, (priority, todoP, abortP))

    def addMap(self, other):
        """
            Takes another PlanMap object and appends whatever the plans are into this plans object.
        """
        # Make sure to extend NOT append these new plans!
        self.plans.extend(other.plans)

        for item in other.plans:
            heapq.heappush(self.workingPlans, item)

    def sort(self):
        """
            Sorts the plans based on priority.
        """
        # Sort by the Priority (first index in tuple)
        self.plans.sort(key=lambda tup: tup[0])

    def registerSkip(self, func):
        """
            Register this function as one to skip execution on, if provided it shouldn't return
            the (func, args) tuple as a result from the getNextTodo function.
        """
        self.skipFunctions.append(func)

    def getNextTodo(self):
        """
        Like an iterator function, it returns each element in the list of plans in order.

        Returns:
            (function, args) : Each todo is returned just how the user first added it
            None             : None is returned when there are no more todo's
        """
        if len(self.workingPlans) == 0:
            return None
        else:
            prio, todo, abt = heapq.heappop(self.workingPlans)
            self.maxPriorityReturned = prio

            # After popping the next plan, check if this is a skipped function.
            # If so, then get the next plan.
            if todo.func in self.skipFunctions:
                return self.getNextTodo()
            else:
                self.abortPlans.append(abt)
                return (todo.func, todo.args)

    def getNextAbort(self):
        """
        Like an iterator function, it returns each element in the list of abort plans in order.

        Returns:
            (function, args) : Each todo is returned just how the user first added it
            None             : None is returned when there are no more todo's
        """
        # Check if this is the first time calling the function
        if not self.startedAbort:
            ###################################################################################
            # If so we generate the abort list right now
            # The reason we do this is that depending on how far along the plan got, we have
            # to abort different things, and some functions will exist in multiple abort plans
            # so we need to call stuff in the proper order
            ###################################################################################

            # Skip the last stage and iterate in reverse order.
            for stage in reversed(self.abortPlans[:-1]):
                # If nothing keep looking
                if not stage:
                    continue

                # This should be a list of Plan() objects
                for a in stage:
                    # See if we are already going to do this Plan()
                    if a not in self.workingAbortPlans:
                        self.workingAbortPlans.append(a)

            # Lastly, set the started flag, so we don't initialize the list again.
            self.startedAbort = True

        if len(self.workingAbortPlans) == 0:
            return None
        else:
            abt = self.workingAbortPlans.pop(0)
            return (abt.func, abt.args)

    def __repr__(self):
        return "<%s %r: %d Plans>" % (self.__class__.__name__, self.name, len(self.plans))

    def __str__(self):
        ret = "%r:\n" % self.name
        for p, todopl, abortpl in self.plans:
            ret += "  %d: %r || %r\n" % (p, todopl, abortpl)
        return ret



####################################################################
# Copyright 2013-2017 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

'''
Contains the functions required to restart chutes properly on power cycle of device.
Checks with pdconfd to make sure it was able to properly bring up all interfaces before
starting chutes.
'''

import json
import time

from paradrop.base.output import out
from paradrop.base.pdutils import timeint
from paradrop.base import constants, settings
from paradrop.core.auth.user import User
from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.core.config.network import reclaimNetworkResources
from paradrop.core.update.update_object import UpdateChute
from paradrop.confd.client import waitSystemUp



FAILURE_WARNING = """
    ************************WARNING************************
    This chute failed to start on a previous attempt for some reason. 
    Perhaps hardware on the device has changed? 
    The chute has been stopped and will need to be started.
    """

def reloadChutes():
    """
    Get update objects to chutes that should be running at startup.

    This function is called to restart any chutes that were running
    prior to the system being restarted.  It waits for pdconfd to
    come up and report whether or not it failed to bring up any of the
    interfaces that existed before the power cycle. If pdconfd indicates
    something failed we then force a stop update in order to bring down
    all interfaces associated with that chute and mark it with a warning.
    If the stop fails we mark the chute with a warning manually and change
    its state to stopped and save to storage this isn't great as it could
    mean our system still has interfaces up associated with that chute.
    If pdconfd doesn't report failure we attempt to start the chute and
    if this fails we trust the abort process to restore the system to a
    consistent state and we manually mark the chute as stopped and add
    a warning to it.

    :returns: (list) A list of UpdateChute objects that should be run
              before accepting new updates.
    """
    if not settings.PDCONFD_ENABLED:
        return []
    chuteStore = ChuteStorage()
    chutes = [ ch for ch in chuteStore.getChuteList() if ch.isRunning() ]

    # Part of restoring the chute to its previously running state is reclaiming
    # IP addresses, interface names, etc. that it had previously.
    for chute in chutes:
        reclaimNetworkResources(chute)

    #We need to make sure confd is up and all interfaces have been brought up properly
    confdup = False
    while not confdup:
        confdInfo = waitSystemUp()
        if confdInfo == None:
            time.sleep(1)
            continue
        confdup = True
        confdInfo = json.loads(confdInfo)

    #Remove any chutes from the restart queue if confd failed to bring up the
    #proper interfaces
    #
    # At this point, we only care about the chute names, not the full objects.
    # We are using sets of chute names for their O(1) membership test and
    # element uniqueness.
    okChutes = set([ch.name for ch in chutes])
    failedChutes = set()
    for iface in confdInfo:
        if iface.get('success') is False:
            failedChuteName = iface.get('comment')
            if failedChuteName == constants.RESERVED_CHUTE_NAME:
                out.warn('Failed to load a system config section')
            elif failedChuteName in okChutes:
                # This was a chute that we are supposed to restart, but one of
                # its config sections failed to load.
                okChutes.remove(failedChuteName)
                failedChutes.add(failedChuteName)
            elif failedChuteName not in failedChutes:
                # In this case, the name in the comment was not something that
                # we recognize from the chute storage.  Maybe the chute storage
                # file was deleted but not the config files, or someone
                # manually modified the config files.  Anyway, we cannot
                # attempt to stop this chute because the object does not exist,
                # but we can give a warning message.
                out.warn('Failed to load config section for '
                         'unrecognized chute: {}'.format(failedChuteName))

    updates = []

    # There was code here that looped over the failedChutes set and explicitly
    # stopped those chutes.  However, if the cause of the failure was
    # transient, those chutes would remain stopped until manually restarted.
    # It seems better to leave them alone so that we try again next time the
    # system reboots.
    #
    # TODO: We should still record the failure somewhere.

    # Only try to restart the chutes that had no problems during pdconf
    # initialization.
    for ch in okChutes:
        update_spec = dict(updateClass='CHUTE', updateType='restart', name=ch,
                tok=timeint(), func=updateStatus,
                user=User.get_internal_user())
        update = UpdateChute(update_spec, reuse_existing=True)
        updates.append(update)

    return updates

def updateStatus(update):
    """
    This function is a callback for the updates we do upon restarting the system.
    It checks whether or not the update completed successfully and if not it
    changes the state of the chute to stopped and adds a warning.
    :param update: The update object containing information about the chute that was created and whether it was successful or not.
    :type update: obj
    :returns: None
    """
    pass



import json

from autobahn.twisted.websocket import WebSocketServerProtocol
from autobahn.twisted.websocket import WebSocketServerFactory
from twisted.internet.task import LoopingCall

from paradrop.base.output import out
from paradrop.core.container.log_provider import LogProvider

class LogSockJSProtocol(WebSocketServerProtocol):
    def __init__(self, factory):
        WebSocketServerProtocol.__init__(self)
        self.factory = factory
        self.loop = LoopingCall(self.check_log)
        self.log_provider = None

    def onOpen(self):
        out.info('sockjs /logs connected')

        self.log_provider = LogProvider(self.factory.chute)
        self.log_provider.attach()
        self.loop.start(1.0)

    def check_log(self):
        logs = self.log_provider.get_logs()
        for log in logs:
            self.sendMessage(json.dumps(log))

    def onClose(self, wasClean, code, reason):
        out.info('sockjs /logs disconnected')

        self.loop.stop()
        self.log_provider.detach()
        self.log_provider = None

class LogSockJSFactory(WebSocketServerFactory):
    def __init__(self, chute):
        WebSocketServerFactory.__init__(self)
        self.chute = chute

    def buildProtocol(self, addr):
        return LogSockJSProtocol(self)



###################################################################
# Copyright 2013-2016 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
This module generates update plans for a snap operation.
"""

from paradrop.base.output import out
from paradrop.core.config import snap
from . import plangraph


def generatePlans(update):
    out.verbose("%r\n" % (update))

    update.plans.addPlans(plangraph.SNAP_INSTALL,
                          (snap.updateSnap, ))



from __future__ import print_function
import json
import pycurl
import re
import six
import urllib

import twisted
from twisted.internet import reactor, threads
from twisted.internet.defer import Deferred, DeferredLock, DeferredSemaphore
from twisted.internet.protocol import Protocol
from twisted.web.client import Agent, FileBodyProducer, HTTPConnectionPool
from twisted.web.http_headers import Headers

from paradrop.base import nexus, settings


class JSONReceiver(Protocol):
    """
    JSON Receiver

    A JSONReceiver object can be used with the twisted HTTP client
    to receive data from a request and provide it to a callback
    function when complete.

    Example (response came from an HTTP request):
    finished = Deferred()
    response.deliverBody(JSONReceiver(finished))
    finished.addCallback(func_that_takes_result)

    Some error conditions will result in the callback firing with a result of
    None.  The receiver needs to check for this.  This seems to occur on 403
    errors where the server does not return any data, but twisted just passes
    us a ResponseDone object the same type as a normal result.
    """
    def __init__(self, response, finished):
        """
        response: a twisted Response object
        finished: a Deferred object
        """
        self.response = response
        self.finished = finished
        self.data = ""

    def dataReceived(self, data):
        """
        internal: handles incoming data.
        """
        self.data += data

    def connectionLost(self, reason):
        """
        internal: handles connection close events.
        """
        if reason.check(twisted.web.client.ResponseDone):
            try:
                result = json.loads(self.data)
            except ValueError:
                result = None

            self.finished.callback(PDServerResponse(self.response, data=result))

        else:
            raise Exception(reason.getErrorMessage())


def urlEncodeParams(data):
    """
    Return data URL-encoded.

    This function specifically handles None and boolean values
    to convert them to JSON-friendly strings (e.g. None -> 'null').
    """
    copy = dict()
    for key, value in six.iteritems(data):
        if value is None:
            copy[key] = 'null'
        elif isinstance(value, bool):
            copy[key] = json.dumps(value)
        else:
            copy[key] = value
    return urllib.urlencode(copy, doseq=True)


class PDServerResponse(object):
    """
    A PDServerResponse object contains the results of a request to pdserver.

    This wraps twisted.web.client.Response (cannot be subclassed) and exposes
    the same variables in addition to a 'data' variables.  The 'data' variable,
    if not None, is the parsed object from the response body.
    """
    def __init__(self, response, data=None):
        self.version = response.version
        self.code = response.code
        self.phrase = response.phrase
        self.headers = response.headers
        self.length = response.length
        self.success = (response.code >= 200 and response.code < 300)
        self.data = data


class HTTPResponse(object):
    def __init__(self, data=None):
        self.version = None
        self.code = None
        self.phrase = None
        self.headers = dict()
        self.length = None
        self.success = False
        self.data = data


class HTTPRequestDriver(object):
    def __init__(self):
        self.headers = {
            "Accept": "application/json",
            "Connection": "keep-alive",
            "Content-Type": "application/json",
            "User-Agent": "ParaDrop/2.5"
        }

    def request(self, method, url, body):
        raise Exception("Not implemented")

    def setHeader(self, key, value):
        self.headers[key] = value


class CurlRequestDriver(HTTPRequestDriver):
    # Shared curl handle.
    # May have problems due to issue #411.
    # https://github.com/pycurl/pycurl/issues/411
    curl = pycurl.Curl()

    # Lock for the access to curl.
    lock = DeferredLock()

    code_pattern = re.compile("(HTTP\S*)\s+(\d+)\s+(.*)")
    header_pattern = re.compile("(\S+): (.*)")

    def __init__(self):
        super(CurlRequestDriver, self).__init__()

        # Buffer for receiving response.
        self.buffer = six.StringIO()

        # Fill in response object.
        self.response = HTTPResponse()

    def receive(self, ignore):
        """
        Receive response from curl and convert it to a response object.
        """
        data = self.buffer.getvalue()

        response = self.response

        # Try to parse the content if it's JSON.
        contentType = response.headers.get('content-type', 'text/html')
        if 'json' in contentType:
            try:
                response.data = json.loads(data)
            except Exception:
                print("Failed to parse JSON")
                print(data)
                response.data = data
        else:
            response.data = data

        response.success = (response.code >= 200 and response.code < 300)
        return response

    def receiveHeaders(self, header_line):
        header_line = header_line.strip()

        match = CurlRequestDriver.code_pattern.match(header_line)
        if match is not None:
            self.response.version = match.group(1)
            self.response.code = int(match.group(2))
            self.response.phrase = match.group(3)
            return

        match = CurlRequestDriver.header_pattern.match(header_line)
        if match is not None:
            key = match.group(1).lower()
            self.response.headers[key] = match.group(2)

    def request(self, method, url, body=None):
        def makeRequest(ignored):
            curl = CurlRequestDriver.curl
            curl.reset()

            curl.setopt(pycurl.URL, url)
            curl.setopt(pycurl.HEADERFUNCTION, self.receiveHeaders)
            curl.setopt(pycurl.WRITEFUNCTION, self.buffer.write)

            curl.setopt(pycurl.CUSTOMREQUEST, method)

            if body is not None:
                curl.setopt(pycurl.POSTFIELDS, body)

            headers = []
            for key, value in six.iteritems(self.headers):
                headers.append("{}: {}".format(key, value))
            curl.setopt(pycurl.HTTPHEADER, headers)

            d = threads.deferToThread(curl.perform)
            d.addCallback(self.receive)
            return d

        def releaseLock(result):
            CurlRequestDriver.lock.release()

            # Forward the result to the next handler.
            return result

        d = CurlRequestDriver.lock.acquire()

        # Make the request once we acquire the semaphore.
        d.addCallback(makeRequest)

        # Release the semaphore regardless of how the request goes.
        d.addBoth(releaseLock)
        return d


class TwistedRequestDriver(HTTPRequestDriver):
    # Using a connection pool enables persistent connections, so we can avoid
    # the connection setup overhead when sending multiple messages to the
    # server.
    pool = HTTPConnectionPool(reactor, persistent=True)

    # Used to control the number of concurrent requests because
    # HTTPConnectionPool does not do that on its own.
    # Discussed here:
    # http://stackoverflow.com/questions/25552432/how-to-make-pooling-http-connection-with-twisted
    sem = DeferredSemaphore(settings.PDSERVER_MAX_CONCURRENT_REQUESTS)

    def receive(self, response):
        """
        Receive response from twisted web client and convert it to a
        PDServerResponse object.
        """
        deferred = Deferred()
        response.deliverBody(JSONReceiver(response, deferred))
        return deferred

    def request(self, method, url, body=None):
        def makeRequest(ignored):
            bodyProducer = None
            if body is not None:
                bodyProducer = FileBodyProducer(six.StringIO(body))

            headers = {}
            for key, value in six.iteritems(self.headers):
                headers[key] = [value]

            agent = Agent(reactor, pool=TwistedRequestDriver.pool)
            d = agent.request(method, url, Headers(headers), bodyProducer)
            d.addCallback(self.receive)
            return d

        def releaseSemaphore(result):
            TwistedRequestDriver.sem.release()

            # Forward the result to the next handler.
            return result

        d = TwistedRequestDriver.sem.acquire()

        # Make the request once we acquire the semaphore.
        d.addCallback(makeRequest)

        # Release the semaphore regardless of how the request goes.
        d.addBoth(releaseSemaphore)
        return d


class PDServerRequest(object):
    """
    Make an HTTP request to pdserver.

    The API is assumed to use application/json for sending and receiving data.
    Authentication is automatically handled here if the router is provisioned.

    We handle missing, invalid, or expired tokens by making the request and
    detecting a 401 (Unauthorized) response.  We request a new token and retry
    the failed request.  We do this at most once and return failure if the
    second attempt returns anything other than 200 (OK).

    PDServerRequest objects are not reusable; create a new one for each
    request.

    URL String Substitutions:
    router_id -> router id

    Example:
    /routers/{router_id}/states -> /routers/halo06/states
    """

    # Auth token (JWT): we will automatically request as needed (for the first
    # request and after expiration) and store the token in memory for future
    # requests.
    token = None

    def __init__(self, path, driver=TwistedRequestDriver, headers={}, setAuthHeader=True):
        self.path = path
        self.driver = driver
        self.headers = headers
        self.setAuthHeader = setAuthHeader
        self.transportRetries = 0

        url = nexus.core.info.pdserver
        if not path.startswith('/'):
            url += '/'
        url += path

        # Perform string substitutions.
        self.url = url.format(router_id=nexus.core.info.pdid)

        self.body = None

    def get(self, **query):
        self.method = 'GET'
        if len(query) > 0:
            self.url += '?' + urlEncodeParams(query)
        d = self.request()
        d.addCallback(self.receiveResponse)
        return d

    def patch(self, *ops):
        """
        Expects a list of operations in jsonpatch format (http://jsonpatch.com/).

        An example operation would be:
        {'op': 'replace', 'path': '/completed', 'value': True}
        """
        self.method = 'PATCH'
        self.body = json.dumps(ops)
        d = self.request()
        d.addCallback(self.receiveResponse)
        return d

    def post(self, **data):
        self.method = 'POST'
        self.body = json.dumps(data)
        d = self.request()
        d.addCallback(self.receiveResponse)
        return d

    def put(self, **data):
        self.method = 'PUT'
        self.body = json.dumps(data)
        d = self.request()
        d.addCallback(self.receiveResponse)
        return d

    def request(self):
        driver = self.driver()
        if self.setAuthHeader and PDServerRequest.token is not None:
            auth = 'Bearer {}'.format(PDServerRequest.token)
            driver.setHeader('Authorization', auth)
        for key, value in six.iteritems(self.headers):
            driver.setHeader(key, value)
        return driver.request(self.method, self.url, self.body)

    def receiveResponse(self, response):
        """
        Intercept the response object, and if it's a 401 authenticate and retry.
        """
        if response.code == 401 and self.setAuthHeader:
            # 401 (Unauthorized) may mean our token is no longer valid.
            # Request a new token and then retry the request.
            #
            # Watch out for infinite recursion here!  If this inner request
            # returns a 401 code, meaning the id/password is invalid, it should
            # not go down this code path again (prevented by check against
            # self.setAuthHeader above).
            authRequest = PDServerRequest('/auth/router', driver=self.driver,
                    setAuthHeader=False)
            d = authRequest.post(id=nexus.core.info.pdid,
                    password=nexus.core.getKey('apitoken'))

            def cbLogin(authResponse):
                if authResponse.success:
                    PDServerRequest.token = authResponse.data.get('token', None)

                    # Retry the original request now that we have a new token.
                    return self.request()

                else:
                    # Our attempt to get a token failed, so give up.
                    return PDServerResponse(response)

            d.addCallback(cbLogin)
            return d

        else:
            return response

    @classmethod
    def getServerInfo(c):
        """
        Return the information needed to send API messages to the server.

        This can be used by an external program (e.g. pdinstall).
        """
        info = {
            'authorization': 'Bearer {}'.format(c.token),
            'router_id': nexus.core.info.pdid,
            'server': nexus.core.info.pdserver
        }
        return info

    @classmethod
    def resetToken(c):
        """
        Reset the auth token, to be called if the router's identity has changed.
        """
        c.token = None


# Initialize pycurl.  Does this do anything?
pycurl.global_init(pycurl.GLOBAL_ALL)


# Set the number of connections that can be kept alive in the connection pool.
# Setting this equal to the size of the semaphore should prevent churn.
TwistedRequestDriver.pool.maxPersistentPerHost = settings.PDSERVER_MAX_CONCURRENT_REQUESTS



import fnmatch

from paradrop.base.output import out
from paradrop.lib.utils import uci

from . import uciutils


def findMatchingInterface(iface_name, interfaces):
    """
    Search an interface list for one matching a given name.

    iface_name can contain shell-style wildcards (* and ?).
    """
    for iface in interfaces:
        if fnmatch.fnmatch(iface['name'], iface_name):
            return iface
    return None


def getOSFirewallRules(update):
    """
    There is a set of default things that must exist just for the chute to function at all, generate those here.

    Stored in key: osFirewallRules
    """
    interfaces = update.cache_get('networkInterfaces')
    if interfaces is None:
        return None

    rules = []

    # Add a zone for each interface.
    for iface in interfaces:
        config = {'type': 'zone'}
        options = {
            'name': iface['externalIntf'],
            'conntrack': True,
            'input': 'REJECT',
            'forward': 'ACCEPT',
            'output': 'ACCEPT',
            'network': [iface['externalIntf']]
        }

        if iface['type'] == 'wan':
            options['masq'] = 1
        else:
            options['masq'] = 0

        # Add the zone section first.
        rules.append((config, options))

        # Then add a forwarding rule for wan type interfaces.
        if iface['type'] == 'wan':
            config = {'type': 'forwarding'}
            options = {
                'src': iface['externalIntf'],
                'dest': 'wan'
            }
            rules.append((config, options))

        # If we are running a DHCP server for the chute, we should allow DHCP
        # and DNS requests to the host.
        if 'dhcp' in iface:
            # Allow DNS requests.
            config = {'type': 'rule'}
            options = {
                'src': iface['externalIntf'],
                'proto': 'udp',
                'dest_port': 53,
                'target': 'ACCEPT'
            }
            rules.append((config, options))

            # Allow DHCP requests.
            config = {'type': 'rule'}
            options = {
                'src': iface['externalIntf'],
                'proto': 'udp',
                'dest_port': 67,
                'target': 'ACCEPT'
            }
            rules.append((config, options))

    update.cache_set('osFirewallRules', rules)


def getDeveloperFirewallRules(update):
    """
    Generate other firewall rules requested by the developer such as redirects.
    The object returned is a list of tuples (config, options).
    """
    interfaces = update.cache_get('networkInterfaces')
    if interfaces is None:
        return None

    rules = list()

    if hasattr(update.new, "firewall"):
        for rule in update.new.firewall:
            if rule['type'] == 'redirect':
                config = {'type': 'redirect'}
                options = {
                    'name': rule['name'],
                    'proto': 'tcpudp'
                }

                from_parts = rule['from'].strip().split(':')
                to_parts = rule['to'].strip().split(':')

                # Do not allow rules that do not pertain to the chute.
                if "@host.lan" in from_parts[0] and "@host.lan" in to_parts[0]:
                    raise Exception("Unable to add firewall rule - "
                                    "dst and src are both outside of chute")

                # From @host.lan means this is a DNAT rule (redirect to the chute).
                if from_parts[0] == "@host.lan":
                    options['target'] = "DNAT"

                    options['src'] = "wan"
                    if len(from_parts) > 1:
                        options['src_dport'] = from_parts[1]

                    # Find the interface specified in the rule, so we can get its
                    # IP address.
                    iface = findMatchingInterface(to_parts[0], interfaces)
                    if iface is None:
                        out.warn("No interface found with name {}\n".format(to_parts[0]))
                        raise Exception("Interface not found")

                    options['dest_ip'] = iface['externalIpaddr']
                    if len(to_parts) > 1:
                        options['dest_port'] = to_parts[1]

                # This is an SNAT rule (redirect from the chute to host network).
                elif to_parts[0] == "@host.lan":
                    options['target'] = "SNAT"

                    # TODO: Implement
                    out.warn("SNAT rules not supported yet")
                    raise Exception("SNAT rules not implemented")

                # Could be forwarding between chute interfaces?
                else:
                    out.warn("Other rules not supported yet")
                    raise Exception("Other rules not implemented")

                rules.append((config, options))

    update.cache_set('developerFirewallRules', rules)


def setOSFirewallRules(update):
    """
    Takes a list of tuples (config, opts) and saves it to the firewall config file.
    """
    uciutils.setConfig(update,
            cacheKeys=['osFirewallRules', 'developerFirewallRules'],
            filepath=uci.getSystemPath("firewall"))


def revert_os_firewall_rules(update):
    uciutils.setConfig(update,
            cacheKeys=['osFirewallRules', 'developerFirewallRules'],
            filepath=uci.getSystemPath("firewall"))



import socket

from http.client import HTTPConnection


class UHTTPConnection(HTTPConnection):
    """
    Subclass of Python library HTTPConnection that uses a unix-domain socket.

    Source: http://7bits.nl/blog/posts/http-on-unix-sockets-with-python
    """

    def __init__(self, path):
        HTTPConnection.__init__(self, 'localhost')
        self.path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self.path)
        self.sock = sock



from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.lib.misc import resopt


def computeResourceAllocation(chutes):
    service_names = []
    service_cpu_fractions = []
    allocation = {}
    for chute in chutes:
        if not chute.isRunning():
            continue

        for service in chute.get_services():
            service_names.append(service.get_container_name())

            cpu_fraction = service.requests.get('cpu-fraction', None)
            service_cpu_fractions.append(cpu_fraction)

    # Use the optimizer to allocate cpu fractions and fill in unspecified
    # (None) values.  The result is a vector that sums to one.
    new_cpu_fractions = resopt.allocate(service_cpu_fractions, total=1.0)

    n = len(service_names)
    for i in range(n):
        # Convert the fraction to an integer.  We multiply by 1024*n so that
        # they all center around 1024, which is what Docker assigns to
        # containers by default.
        cpu_shares = int(round(new_cpu_fractions[i] * 1024 * n))

        # Keep it above 2.  Docker treats 0 and 1 as special values.
        cpu_shares = max(cpu_shares, 2)

        name = service_names[i]
        allocation[name] = {
            'cpu_shares': cpu_shares
        }

    return allocation


def getResourceAllocation(update):
    """
    Allocate compute resources for chutes.

    Sets cache variables "newResourceAllocation" and "oldResourceAllocation".
    """
    chuteStore = ChuteStorage()
    chutes = chuteStore.getChuteList()

    old_allocation = computeResourceAllocation(chutes)

    # Insert the new chute into the list and replace the old version if it
    # existed.
    chutes.insert(0, update.new)
    for i in range(1, len(chutes)):
        if chutes[i].name == update.new.name:
            del chutes[i]
            break

    new_allocation = computeResourceAllocation(chutes)

    update.cache_set('newResourceAllocation', new_allocation)
    update.cache_set('oldResourceAllocation', old_allocation)



###################################################################
# Copyright 2013-2014 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import socket
import struct

from paradrop.lib.utils import pdos


def isIpValid(ipaddr):
    """Return True if Valid, otherwise False."""
    try:
        socket.inet_aton(ipaddr)
        return True
    except:
        return False
    

def isIpAvailable(ipaddr, chuteStor, name):
    """
    Make sure this IP address is available. 

    Checks the IP addresses of all zones on all other chutes, makes sure
    subnets are not the same.
    """
    chList = chuteStor.getChuteList()
    for ch in chList:
        # Only look at other chutes
        if (name != ch.name):
            otherIPs = ch.getChuteIPs()
            if ipaddr in otherIPs:
                return False
    
    return True


def isWifiSSIDAvailable(ssid, chuteStor, name):
    """Make sure this SSID is available."""
    chList = chuteStor.getChuteList()
    for ch in chList:
        # Only look at other chutes
        if (name != ch.name):
            otherSSIDs = ch.getChuteSSIDs()    
            #print("chute: %s other SSIDs: %s" % (ch, otherSSIDs))
            for o in otherSSIDs:        
                if (o == ssid):
                    return False

    return True


def isStaticIpAvailable(ipaddr, chuteStor, name):
    """
    Make sure this static IP address is available. 

    Checks the IP addresses of all zones on all other chutes,
    makes sure not equal."""
    chList = chuteStor.getChuteList()
    for ch in chList:
        # Only look at other chutes
        if (name != ch.name):
            otherStaticIPs = ch.getChuteStaticIPs()
            if ipaddr in otherStaticIPs:
                return False 
    
    return True


def checkPhyExists(radioid):
    """Check if this chute exists at all, a directory /sys/class/ieee80211/phyX must exist."""
    #DFW: used to be this, but when netns runs this doesn't exist, so we switched to using the debug sysfs '/sys/class/ieee80211/phy%d' % radioid
    return pdos.exists('/sys/kernel/debug/ieee80211/phy%d' % radioid)


def incIpaddr(ipaddr, inc=1):
    """
        Takes a quad dot format IP address string and adds the @inc value to it by converting it to a number.

        Returns:
            Incremented quad dot IP string or None if error
    """
    try:
        val = struct.unpack("!I", socket.inet_aton(ipaddr))[0]
        val += inc
        return socket.inet_ntoa(struct.pack('!I', val))
    except Exception:
        return None


def maxIpaddr(ipaddr, netmask):
    """
        Takes a quad dot format IP address string and makes it the largest valid value still in the same subnet.

        Returns:
            Max quad dot IP string or None if error
    """
    try:
        val = struct.unpack("!I", socket.inet_aton(ipaddr))[0]
        nm = struct.unpack("!I", socket.inet_aton(netmask))[0]
        inc = struct.unpack("!I", socket.inet_aton("0.0.0.254"))[0]
        val &= nm
        val |= inc
        return socket.inet_ntoa(struct.pack('!I', val))
    except Exception:
        return None


def getSubnet(ipaddr, netmask):
    try:
        val1 = struct.unpack("!I", socket.inet_aton(ipaddr))[0]
        nm = struct.unpack("!I", socket.inet_aton(netmask))[0]
        res = val1 & nm
        return socket.inet_ntoa(struct.pack('!I', res))
    except Exception:
        return None


def getInternalIntfList(ch):
    """
        Takes a chute object and uses the key:networkInterfaces to return a list of the internal
        network interfaces that will exist in the chute (e.g., eth0, eth1, ...)

        Returns:
            A list of interface names
            None if networkInterfaces doesn't exist or there is an error
    """
    intfs = ch.getCache('networkInterfaces')
    if(not intfs):
        return None
    l = []
    for i in intfs:
        l.append(i['internalIntf'])

    return l


def getGatewayIntf(ch):
    """
        Looks at the key:networkInterfaces for the chute and determines what the gateway should be
        including the IP address and the internal interface name.

        Returns:
            A tuple (gatewayIP, gatewayInterface)
            None if networkInterfaces doesn't exist or there is an error
    """
    intfs = ch.getCache('networkInterfaces')
    if(not intfs):
        return (None, None)

    for i in intfs:
        if(i['type'] == 'wan'):
            return (i['externalIpaddr'], i['internalIntf'])

    return (None, None)


def getWANIntf(ch):
    """
        Looks at the key:networkInterfaces for the chute and finds the WAN interface.

        Returns:
            The dict from networkInterfaces
            None
    """
    intfs = ch.getCache('networkInterfaces')
    if(not intfs):
        return None
    
    for i in intfs:
        if(i['type'] == 'wan'):
            return i
    return None



import attr
import six


@attr.s
class User(object):
    role_priorities = {
        "guest": 0,
        "user": 1,
        "trusted": 2,
        "admin": 3,
        "owner": 4
    }

    # User attributes
    name = attr.ib()
    domain = attr.ib()
    role = attr.ib(default="user")

    def __eq__(self, other):
        if isinstance(other, User):
            return (self.name == other.name and self.domain == other.domain)
        elif isinstance(other, six.string_types):
            return self.name == other
        else:
            return False

    @classmethod
    def get_internal_user(user):
        """
        Create a User object for internal usage.

        This user will have unlimited privileges.
        """
        return user("__paradrop__", "localhost", role="owner")

    @staticmethod
    def get_highest_role(roles):
        x = zip((User.role_priorities[r] for r in roles), roles)
        x.sort(reverse=True)
        return x[0][1]



'''
APIs for developers to check whether Airshark feature is available or not
'''
import json
from klein import Klein

from . import cors


class AirsharkApi(object):
    routes = Klein()

    def __init__(self, airshark_manager):
        self.airshark_manager = airshark_manager

    @routes.route('/status')
    def status(self, request):
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')
        hardware_ready, software_ready, airshark_running \
            = self.airshark_manager.status()
        data = dict()
        data['hardware_ready'] = hardware_ready
        data['software_ready'] = software_ready
        data['airshark_running'] = airshark_running
        return json.dumps(data)
